import os
import sys

from aesara_theano_fallback import aesara as theano
import argparse
from astropy.units import UnitsWarning
from astropy.stats import mad_std
from celerite2.backprop import LinAlgError
from configparser import ConfigParser
from datetime import datetime
import gc
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import shutil
from timeit import default_timer as timer
import warnings

from alderaan.constants import *
from alderaan.ephemeris import Ephemeris
from alderaan.litecurve import LiteCurve
from alderaan.planet import Planet
from alderaan.modules.detrend import GaussianProcessDetrender
from alderaan.modules.omc import OMC
from alderaan.modules.transit_model import ShapeTransitModel, TTimeTransitModel
from alderaan.modules.quality_control import QualityControl
from alderaan.modules.quicklook import plot_litecurve, plot_omc, dynesty_cornerplot, dynesty_runplot, dynesty_traceplot
from alderaan.utils.io import resolve_config_path, parse_koi_catalog, parse_holczer16_catalog, copy_input_target_catalog


def initialize_pipeline():
    # flush buffer
    sys.stdout.flush()
    sys.stderr.flush()

    # filter warnings
    warnings.simplefilter('always', UserWarning)
    warnings.filterwarnings(
        action='ignore', category=UnitsWarning, module='astropy'
    )

    # start timer
    global_start_time = timer()

    return global_start_time


def cleanup():
    sys.stdout.flush()
    sys.stderr.flush()
    plt.close('all')
    gc.collect()


def main():
    # initialize program
    global_start_time = initialize_pipeline()

    print("")
    print("+" * shutil.get_terminal_size().columns)
    print("ALDERAAN Pipeline")
    print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
    print("+" * shutil.get_terminal_size().columns)
    print("")

    # read inputs from config
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--mission', required=True, type=str, help='Observing mission')
    parser.add_argument('-t', '--target', required=True, type=str, help='Target ID for star')
    parser.add_argument('-c', '--config', required=True, type=str, help='Path to config file')
    args = parser.parse_args()

    config = ConfigParser()
    config.read(args.config)

    alderaan_base_path = Path(__file__).resolve().parents[2]
    for key, value in config["PATHS"].items():
        config['PATHS'][key] = resolve_config_path(config['PATHS'][key], alderaan_base_path)

    mission = args.mission
    target = args.target
    run_id = config['RUN']['run_id']

    data_dir =  config['PATHS']['data_dir']
    outputs_dir = config['PATHS']['outputs_dir']
    catalog_dir = config['PATHS']['catalog_dir']

    catalog_csv = os.path.join(catalog_dir, str(config['ARGS']['catalog_csv']))

    print("")
    print(f"   MISSION : {mission}")
    print(f"   TARGET  : {target}")
    print(f"   RUN ID  : {run_id}")
    print("")
    print(f"   Data directory    : {data_dir}")
    print(f"   Config file       : {args.config}")
    print(f"   Input catalog     : {os.path.basename(catalog_csv)}")
    print("")
    print(f"   theano cache : {theano.config.compiledir}")
    print("")

    # build directory structure
    os.makedirs(outputs_dir, exist_ok=True)

    results_dir = os.path.join(outputs_dir, 'results', run_id, target)
    os.makedirs(results_dir, exist_ok=True)

    quicklook_dir = os.path.join(outputs_dir, 'quicklook', run_id, target)
    os.makedirs(quicklook_dir, exist_ok=True)

    # copy input catalog into results directory
    catalog_csv_copy = os.path.join(outputs_dir, 'results', run_id, f'{run_id}.csv')
    copy_input_target_catalog(catalog_csv, catalog_csv_copy)


    # ######### #
    # I/O Block #
    # ######### #

    print('\n\nI/O BLOCK\n')

    # load KOI catalog
    catalog = parse_koi_catalog(catalog_csv, target)

    assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

    NPL = int(catalog.npl[0])
    koi_id = catalog.koi_id[0]
    kic_id = int(catalog.kic_id[0])

    # load lightcurves
    #litecurve_master = LiteCurve(data_dir, kic_id, 'long cadence', data_source='Kepler PDCSAP')
    litecurve_master = LiteCurve().from_kplr_pdcsap(data_dir, kic_id, 'long cadence', visits=2)

    t_min = litecurve_master.time.min()
    t_max = litecurve_master.time.max()
    if t_min < 0:
        raise ValueError("Lightcurve has negative timestamps...this will cause problems")

    # split litecurves by quarter
    litecurves = litecurve_master.split_visits()

    for j, litecurve in enumerate(litecurves):
        assert len(np.unique(litecurve.visit)) == 1, "expected one quarter per litecurve"
        assert len(np.unique(litecurve.obsmode)) == 1, "expected one obsmode per litecurve"

    print(f"{len(litecurves)} litecurves loaded for {target}")

    # initialize planets (catch no ephemeris warning)
    with warnings.catch_warnings(record=True) as catch:
        warnings.simplefilter('always', category=UserWarning)
        planets = [None]*NPL
        for n in range(NPL):
            planets[n] = Planet(catalog, target, n)

    print(f"\n{NPL} planets loaded for {target}")
    print([np.round(p.period,6) for p in planets])

    # update planet ephemerides
    for n, p in enumerate(planets):
        if p.ephemeris is None:
            _ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
            planets[n] = p.update_ephemeris(_ephemeris)

    # load Holczer+2016 catalog
    filepath = os.path.join(catalog_dir, 'holczer_2016_kepler_ttvs.txt')
    holczer_ephemerides = parse_holczer16_catalog(filepath, koi_id, NPL)

    print(f"\n{len(holczer_ephemerides)} ephemerides found in Holczer+2016")

    # match Holczer ephemerides to Planets
    count = 0

    for n, p in enumerate(planets):
        for ephem in holczer_ephemerides:
            match = np.isclose(ephem.period, p.period, rtol=0.01, atol=p.duration)

            if match:
                print(f"  Planet {n} : {p.period:.6f} --> {ephem.period:.6f}")
                planets[n] = p.update_ephemeris(ephem)
                count += 1

    print(f"{count} matching ephemerides found ({len(holczer_ephemerides)} expected)")

    # quicklook litecurve
    filepath = os.path.join(quicklook_dir, f"{target}_litecurve_raw.png")
    _ = plot_litecurve(litecurve_master, target, planets, filepath)

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ######### #
    # OMC Block #
    # ######### #

    print('\n\nOMC BLOCK (initialization)\n')
    print("regularizing ephemerides")

    # initialize OMC object for each planet
    omc_list = []
    for n, p in enumerate(planets):
        omc_list.append(OMC(p.ephemeris))

    # fit a regularized model
    for n, p in enumerate(planets):
        omc = omc_list[n]
        npts = np.sum(omc.quality)

        _period = np.copy(p.period)

        # Matern-3/2 model | don't use GP on very noisy data
        if (npts >= 8) & (np.median(omc.yerr) <= 0.5 * mad_std(omc.yobs)):
            with warnings.catch_warnings(record=True) as catch:
                warnings.simplefilter('always', category=RuntimeWarning)
                trace = omc.sample(omc.matern32_model())

        # Polynomial model | require 2^N transits
        else:
            polyorder = np.max([1, np.min([3, int(np.log2(npts))-1])])
            with warnings.catch_warnings(record=True) as catch:
                warnings.simplefilter('always', category=RuntimeWarning)
                trace = omc.sample(omc.poly_model(polyorder))

        if len(catch) > 0:
            print(f"{len(catch)} RuntimeWarnings caught during sampling")

        # update ephemeris
        omc.ymod = np.nanmedian(trace['pred'], 0)
        omc_list[n] = omc

        p.ephemeris = p.ephemeris.update_from_omc(omc)
        p.ephemeris = p.ephemeris.interpolate('spline', full=True)
        planets[n] = p.update_ephemeris(p.ephemeris)

        # make quicklook plot
        filepath = os.path.join(quicklook_dir, f"{target}_omc_initial.png")
        _ = plot_omc(omc_list, target, filepath)

        print(f"Planet {n} : {_period:.6f} --> {planets[n].period:.6f}")

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ################ #
    # DETRENDING BLOCK #
    # ################ #

    print('\n\nDETRENDING BLOCK (1st pass)\n')

    # initialize detrenders
    detrenders = []
    for j, litecurve in enumerate(litecurves):
        detrenders.append(GaussianProcessDetrender(litecurve, planets))

    # clip outliers
    for j, detrender in enumerate(detrenders):    
        mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
        
        npts_initial = len(detrender.litecurve.time)

        detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=5, mask=mask)
        detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=1000, mask=None)
        
        npts_final = len(detrender.litecurve.time)

        print(f"  Quarter {detrender.litecurve.visit[0]} : {npts_initial-npts_final} outliers rejected")

    # estimate oscillation periods
    oscillation_periods = np.zeros(len(detrenders))
    for j, detrender in enumerate(detrenders):
        obsmode = detrender.litecurve.obsmode[0]

        if obsmode == 'short cadence':
            min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
        elif obsmode == 'long cadence':
            min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
        else:
            raise ValueError(f"unsuported obsmode: {obsmode}")

        oscillation_periods[j] = detrender.estimate_oscillation_period(min_period=min_period)

    osc_per_mu = np.nanmedian(oscillation_periods)
    osc_per_sd = np.max([mad_std(oscillation_periods, ignore_nan=True), 0.1*osc_per_mu])

    print(f"\nNominal stellar oscillation period = {osc_per_mu:.1f} +/- {osc_per_sd:.1f} days")

    # detrend the litecurves
    for j, detrender in enumerate(detrenders):
        print(f"\nDetrending {j+1} of {len(detrenders)} litecurves", flush=True)
        
        # set detrender arguments based on observing mode
        obsmode = detrender.litecurve.obsmode[0]

        if obsmode == 'short cadence':
            min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
            gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_scit * 5 / 2), 91])
            jump_tolerance = 5.0
        elif obsmode == 'long cadence':
            min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
            gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_lcit * 5 / 2), 13])
            jump_tolerance = 5.0
        else:
            raise ValueError(f"unsuported obsmode: {obsmode}")
        
        # make transit mask
        transit_mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
        
        # call detrender.detrend(), using successively simpler models as fallbacks
        try:
            litecurves[j] = detrender.detrend(
                'RotationTerm',
                np.nanmedian(oscillation_periods),
                min_period,
                transit_mask=transit_mask,
                gap_tolerance=gap_tolerance,
                jump_tolerance=jump_tolerance,
                correct_ramp=True,
                return_trend=False, 
                progressbar=False
            )
        except LinAlgError:
            warnings.warn(
                "Initial detrending failed...attempting to refit without exponential ramp component"
            )

            try:
                litecurves[j] = detrender.detrend(
                    'RotationTerm',
                    np.nanmedian(oscillation_periods),
                    min_period,
                    transit_mask=transit_mask,
                    gap_tolerance=gap_tolerance,
                    jump_tolerance=jump_tolerance,
                    correct_ramp=False,
                    return_trend=False, 
                    progressbar=False
                )
            except LinAlgError:
                warnings.warn(
                    "Detrending with RotationTerm failed...attempting to detrend with SHOTerm"
                )
                litecurves[j] = detrender.detrend(
                    'SHOTerm',
                    np.nanmedian(oscillation_periods),
                    min_period,
                    transit_mask=transit_mask,
                    gap_tolerance=gap_tolerance,
                    jump_tolerance=jump_tolerance,
                    correct_ramp=False,
                    return_trend=False, 
                    progressbar=False
                )

    # recombine litecurves
    litecurve = LiteCurve().from_list(litecurves)

    # quicklook litecurve
    filepath = os.path.join(quicklook_dir, f"{target}_litecurve_detrended.png")
    _ = plot_litecurve(litecurve, target, planets, filepath)

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ##################### #
    # QUALITY CONTROL BLOCK #
    # ##################### #

    print("\n\nQUALITY CONTROL BLOCK")

    qc = QualityControl(litecurve, planets)

    # check for transits with poor photometric coverage
    with warnings.catch_warnings(record=True) as catch:
        warnings.simplefilter('always', category=RuntimeWarning)
        coverage = qc.check_coverage()

    # check for transits with unusually high noise
    with warnings.catch_warnings(record=True) as catch:
        warnings.simplefilter('always', category=RuntimeWarning)
        good_rms = qc.check_rms(rel_size=3.0, abs_size=2/24, sigma_cut=5.0)

    for n, p in enumerate(planets):
        print(f"\n  Planet {n}:")

        assert len(coverage[n]) == len(p.ephemeris.ttime)
        assert len(good_rms[n]) == len(p.ephemeris.ttime)

        _nbad = np.sum(~coverage[n])
        _ntot = len(coverage[n])
        print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for insufficent photometric coverage")

        _nbad = np.sum(~good_rms[n] & coverage[n])
        _ntot = len(good_rms[n])
        print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for high photometric noise")
        
        planets[n].ephemeris.quality = coverage[n] & good_rms[n]

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ################### #
    # TRANSIT MODEL BLOCK #
    # ################### #

    print('\n\nTRANSIT MODEL BLOCK\n')

    limbdark = [catalog.limbdark_1[0], catalog.limbdark_2[0]]
    transitmodel = ShapeTransitModel(litecurve, planets, limbdark)

    print("Supersample factor")
    for obsmode in transitmodel.unique_obsmodes:
        print(f"  {obsmode} : {transitmodel._obsmode_to_supersample(obsmode)}")

    print("\nFitting initial transit model")
    theta = transitmodel.optimize()
    planets = transitmodel.update_planet_parameters(theta)
    limbdark = transitmodel.update_limbdark_parameters(theta)

    print("\nFitting independent transit times")
    ttvmodel = TTimeTransitModel(litecurve, planets, limbdark)

    for n, p in enumerate(planets):
        print(f"\nPlanet {n} : fitting {np.sum(p.ephemeris.quality)} transit times")

        ttime_new, ttime_err_new = ttvmodel.mazeh13_holczer16_method(n, quicklook_dir=quicklook_dir)

        assert len(ttime_new) == len(ttime_err_new)
        assert len(ttime_new) == len(p.ephemeris.ttime)

        _nfit = np.sum(~np.isnan(ttime_new))
        _ntot = np.sum(p.ephemeris.quality)

        print(f"Planet {n} : {_nfit} of {_ntot} transit times ({_nfit / _ntot * 100:.1f}%) fit successfully")

    print("\nSampling with DynamicNestedSampler")
    results = transitmodel.sample(progress_every=10)

    filepath = os.path.join(quicklook_dir, f'{koi_id}_dynesty_runplot.png')
    fig, ax = dynesty_runplot(results, koi_id, filepath=filepath)

    for n, p in enumerate(transitmodel.planets):
        filepath = os.path.join(quicklook_dir, f'{koi_id}_dynesty_traceplot_{n:02d}.png')
        fig, ax = dynesty_traceplot(results, koi_id, n, filepath=filepath)

        filepath=os.path.join(quicklook_dir, f'{koi_id}_dynesty_cornerplot_{n:02d}.png')
        fig, ax = dynesty_cornerplot(results, koi_id, n, filepath=filepath, interactive=True)


if __name__ == '__main__':
    main()


################################
# - Combine Umbrella Samples - #
################################

# This script combines MCMC samples from different umbrellas into a single chain
# See Gilbert 2022 for details

import os
import sys
import glob
import shutil
import warnings
from datetime import datetime
from timeit import default_timer as timer

print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Posterior chain inspection")
print("Initialized {0}".format(datetime.now().strftime("%d-%b-%Y at %H:%M:%S")))
print("+" * shutil.get_terminal_size().columns)
print("")

# start program timer
global_start_time = timer()

# Parse inputs
import argparse
import matplotlib as mpl

try:
    parser = argparse.ArgumentParser(
        description="Inputs for ALDERAAN transit fiting pipeline"
    )
    parser.add_argument(
        "--mission",
        default=None,
        type=str,
        required=True,
        help="Mission name; can be 'Kepler' or 'Simulated'",
    )
    parser.add_argument(
        "--target",
        default=None,
        type=str,
        required=True,
        help="Target name; format should be K00000 or S00000",
    )
    parser.add_argument(
        "--planet_no",
        default=None,
        type=int,
        required=True,
        help="ALDERAAN zero-indexed planet identifier (i.e. *NOT* KOI_ID)",
    )
    parser.add_argument(
        "--root_dir",
        default=None,
        type=str,
        required=True,
        help="Root directory for system",
    )
    parser.add_argument(
        "--project_dir",
        default=None,
        type=str,
        required=True,
        help="Project directory for accessing lightcurve data and saving outputs",
    )
    parser.add_argument(
        "--interactive",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable interactive plotting; by default matplotlib backend will be set to 'Agg'",
    )

    args = parser.parse_args()
    MISSION = args.mission
    TARGET = args.target
    PLANET_NO = args.planet_no
    ROOT_DIR = args.root_dir
    PROJECT_DIR = ROOT_DIR + args.project_dir

    # set plotting backend
    if args.interactive == False:
        mpl.use("agg")

except:
    pass

# Set environment variables
sys.path.append(PROJECT_DIR)

# Build directory structure
FIGURE_DIR = PROJECT_DIR + "Figures/" + TARGET + "/"
TRACE_DIR = PROJECT_DIR + "Traces/" + TARGET + "/"
QUICK_TTV_DIR = PROJECT_DIR + "QuickTTVs/" + TARGET + "/"
DLC_DIR = PROJECT_DIR + "Detrended_lightcurves/" + TARGET + "/"
NOISE_DIR = PROJECT_DIR + "Noise_models/" + TARGET + "/"

# check if all the output directories exist and if not, create them
if os.path.exists(FIGURE_DIR) == False:
    os.mkdir(FIGURE_DIR)

if os.path.exists(TRACE_DIR) == False:
    os.mkdir(TRACE_DIR)

if os.path.exists(QUICK_TTV_DIR) == False:
    os.mkdir(QUICK_TTV_DIR)

if os.path.exists(DLC_DIR) == False:
    os.mkdir(DLC_DIR)

if os.path.exists(NOISE_DIR) == False:
    os.mkdir(NOISE_DIR)


# Import packages
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import arviz as az
import astropy
from astropy.io import fits
import corner

from alderaan.constants import *
from alderaan.utils import *
from alderaan.sampling import *
import alderaan.umbrella as umb

# flush buffer to avoid mixed outputs from progressbar
sys.stdout.flush()

# turn off FutureWarnings
warnings.filterwarnings("ignore", category=FutureWarning)

# check for interactive matplotlib backends
if np.any(np.array(["agg", "png", "svg", "pdf", "ps"]) == mpl.get_backend()):
    iplot = False
else:
    iplot = True


# Load MCMC posterior chains
# These can be generated by running "fit_transit_shape_independently.py"
files = glob.glob(TRACE_DIR + TARGET + "*{0}*".format(str(PLANET_NO).zfill(2)))
raw_data = {}

for i, f in enumerate(files):
    if ("N_indep" in f) or ("T_indep" in f) or ("G_indep" in f):
        with fits.open(f) as HDUList:
            ext = HDUList[0].header["UMBRELLA"]
            raw_data[ext] = pd.DataFrame(HDUList[1].data)

            # FITS stores big-endian data; we want little-endian
            for k in raw_data[ext].keys():
                if raw_data[ext][k].dtype == ">f8":
                    raw_data[ext][k] = raw_data[ext][k].astype("float")
                elif raw_data[ext][k].dtype == ">i8":
                    raw_data[ext][k] = raw_data[ext][k].astype("int")
                else:
                    warnings.warn(
                        "Expected all dtypes to be '>f8' or '>i8', but found '{0}'".format(
                            raw_data[ext][k].dtype
                        )
                    )

            # calculate convenience quantities
            for k in raw_data.keys():
                raw_data[k]["ROR"] = np.exp(raw_data[k]["LN_ROR"])
                raw_data[k]["DUR14"] = np.exp(raw_data[k]["LN_DUR14"])


# Visually inspect chains
fig, ax = plt.subplots(1, 1, figsize=(8, 5))
bins = np.linspace(0, 2, 21)

g = raw_data["T"].GAMMA

y, x, _ = ax.hist(
    g, bins=bins, density=True, histtype="step", color="C{0}".format(PLANET_NO), lw=2
)

ax.axvline(1, ls=":", color="k")
ax.text(
    0.95, y.max() * 1.13, r"$\leftarrow$ grazing", va="top", ha="right", fontsize=14
)
ax.text(
    1.05, y.max() * 1.13, r"non-grazing $\rightarrow$", va="top", ha="left", fontsize=14
)
ax.set_xlabel(r"$\gamma$", fontsize=20)
ax.set_ylim(0, y.max() * 1.15)
ax.set_yticks([])
ax.set_title(TARGET + "-{0}".format(PLANET_NO), fontsize=20)
plt.savefig(
    FIGURE_DIR + TARGET + "_transition_gamma_{0}.png".format(PLANET_NO),
    bbox_inches="tight",
)

if iplot:
    plt.show()
else:
    plt.close()

print(
    "{0} out of {1} transition samples are consistent with a grazing transit".format(
        np.sum(g < 1), len(g)
    )
)


# Calculate umbrella weights
gN = raw_data["N"].GAMMA
gT = raw_data["T"].GAMMA
gG = raw_data["G"].GAMMA

# Case 1: nongrazing transit near stellar center
if (np.sum(gT < 1) == 0) & (np.sum(gN < 2) == 0):
    trajectory = "N_only"
    psi_fxns = [umb.psi_N]
    psi_names = ["N"]

# Case 2: nongrazing transit near stellar limb
elif (np.sum(gT < 1) == 0) & (np.sum(gN < 2) != 0):
    trajectory = "N_or_T"
    psi_fxns = [umb.psi_N, umb.psi_T]
    psi_names = ["N", "T"]

# Case 3: ambiguous
elif (np.sum(gT < 1) != 0) & (np.sum(gT > 1) != 0):
    trajectory = "any"
    psi_fxns = [umb.psi_N, umb.psi_T, umb.psi_G]
    psi_names = ["N", "T", "G"]

# Case 4: weakly grazing
elif (np.sum(gT < 1) != 0) & (np.sum(gG > 0) != 0):
    trajectory = "T_or_G"
    psi_fxns = [umb.psi_T, umb.psi_G]
    psi_names = ["T", "G"]

# Case 5: strongly grazing
elif (np.sum(gT < 1) != 0) & (np.sum(gG > 0) == 0):
    trajectory = "G_only"
    psi_fxns = [umb.psi_G]
    psi_names = ["G"]

if len(psi_names) != len(psi_fxns):
    raise ValueError("mismatch between specified number of umbrellas")

plt.figure()
x = np.linspace(-1, 5, 1000)
for i, psi in enumerate(psi_fxns):
    plt.plot(x, psi(x) / psi(x).max(), lw=2.5, label=psi_names[i])
plt.xlabel("$\gamma$", fontsize=20)
plt.ylabel(r"$\psi(\gamma)$", fontsize=20)
plt.legend(fontsize=12)
if iplot:
    plt.show()
else:
    plt.close()


# Post-processed chains will only store core physical parameters
keys = "PERIOD T0 ROR IMPACT DUR14 LD_U1 LD_U2".split()

if len(psi_fxns) == 1:
    pname = psi_names[0]
    uweights = 1 / psi_fxns[0](raw_data[pname].GAMMA)

    samples = raw_data[pname][keys]
    weights = uweights / np.sum(uweights)

elif len(psi_fxns) > 1:
    coords = []

    for i, pname in enumerate(psi_names):
        coords.append(np.array(raw_data[pname].GAMMA))

    # starting estimate of z using single-bin WHAM
    z = np.zeros(len(psi_fxns))

    for i, psi in enumerate(psi_fxns):
        z[i] = np.mean(psi(coords[i]))

    z /= np.sum(z)

    print("Starting estimate for z:  ", z.round(4))

    # first iteration of EMUS
    F = umb.F_iter(z, psi_fxns, coords)
    z = umb.z_iter(F)

    print("First EMUS iteration:  ", z.round(4))
    print(z)

    # iterate until convergence
    for n in range(100):
        # for numerical stability
        z_old = np.copy(z)
        z_old[z_old < 1e-100] = 1e-100

        # Eq. 4 & 7 again
        F = umb.F_iter(z, psi_fxns, coords)
        z = umb.z_iter(F)

        # check if we have converged
        if np.max(np.abs(z - z_old) / z_old) < 1e-10:
            break

    z_emus = np.copy(z)

    print("Final converged z:  ", z.round(4))

    # compute weights as in eq 4 in Matthews et al.
    uweights = []
    for i in range(3):
        wd = 0.0

        for j in range(3):
            wd += psi_fxns[j](coords[i]) / z[j]

        uweights.append(1.0 / wd)

    # concatenate samples and track weights for ease of use
    samples = pd.concat(
        [raw_data["N"], raw_data["T"], raw_data["G"]], ignore_index=True
    )[keys]
    weights = np.hstack(uweights) / np.sum(np.hstack(uweights))

# Estimate KDE bandwidths
bandwidths = pd.DataFrame()

for k in samples.keys():
    print(k)

    bandwidths[k] = [1.0 * get_bw(samples[k], weights)]

# Generate synthetic samples with PDF-Oversampling (Gao+ 2014)
synth_samps = generate_synthetic_samples(
    samples.values, bandwidths.values, n_up=len(samples), weights=weights
)
synth_samps = pd.DataFrame(synth_samps, columns=samples.keys())

# Flag 10-sigma outliers
out = np.zeros(len(synth_samps.PERIOD), dtype="bool")

for k in synth_samps.keys():
    out += astropy.stats.sigma_clip(
        synth_samps[k], cenfunc="median", stdfunc="mad_std", sigma=10
    ).mask

# Make corner plot
fig = corner.corner(synth_samps[~out], color="C{0}".format(PLANET_NO))
plt.savefig(
    FIGURE_DIR + TARGET + "_corner_{0}.png".format(PLANET_NO), bbox_inches="tight"
)

if iplot:
    plt.show()
else:
    plt.close()


# Write combined MCMC samples to .fits file
print("\nWriting MCMC samples to .fits\n")

# set up header info
data_info = [
    # ['VARIABLE', 'UNITS',        'DESCRIPTION',                           'PRIOR'],
    ["PERIOD", "days", "Orbital period", "Normal"],
    ["T0", "days", "Mid-point of first transit", "Normal"],
    ["ROR", "dimensionless", "Planet-to-star radius ratio", "Log-Uniform"],
    ["IMPACT", "dimensionless", "Transit impact parameter", "Uniform"],
    ["DUR14", "days", "Transit duration: 1st-4th contact", "Log-Uniform"],
    [
        "LD_U1",
        "dimensionless",
        "1st quadratic stellar limbdark coeff",
        "Kipping (2013)",
    ],
    [
        "LD_U2",
        "dimensionless",
        "2nd quadratic stellar limbdark coeff",
        "Kipping (2013)",
    ],
]

# build primary HDU and header
hdu = fits.PrimaryHDU()
hdu.header["TARGET"] = (TARGET + "-" + str(PLANET_NO).zfill(2), "Target name")
hdu.header["PHOTOSRC"] = (MISSION, "Source of photometry")
hdu.header["DATETIME"] = (
    datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    "Date and time of analysis",
)
hdu.header[
    "COMMENT"
] = "ALDERAAN single planet transit fit -- Combined Synthetic Samples"

f_name = PROJECT_DIR + "Traces/{0}/{0}-{1}_CSS_indep.fits".format(
    TARGET, str(PLANET_NO).zfill(2)
)
hdu.writeto(f_name, overwrite=True)

# Store data from dataframes into a single FITS file
with fits.open(f_name) as hduL:
    # Store trace data and summary data to HDUs for each model
    rec_data = synth_samps.to_records(index=False)
    hdu_data = fits.BinTableHDU(data=rec_data, name="SAMPLES")

    hduL.append(hdu_data)

    data_head = hduL["SAMPLES"].header
    for i, dinf in enumerate(data_info):
        data_head[dinf[0]] = (dinf[1], dinf[2])

    hduL.writeto(f_name, overwrite=True)


# Exit program
print("")
print("+" * shutil.get_terminal_size().columns)
print(
    "Umbrella sample processing complete {0}".format(
        datetime.now().strftime("%d-%b-%Y at %H:%M:%S")
    )
)
print("Total runtime = %.1f min" % ((timer() - global_start_time) / 60))
print("+" * shutil.get_terminal_size().columns)



__all__ = ['Planet']

import numpy as np
from alderaan.ephemeris import Ephemeris
import warnings


class Planet:
    """Planet
    """
    def __init__(self, 
                 catalog,
                 koi_id,
                 planet_no, 
                 ephemeris=None
                ):
        # read transit parameters from pandas dataframe
        self = self._from_dataframe(catalog, koi_id, planet_no)

        # set up ephemeris
        if ephemeris is not None:
            self = self.update_ephemeris(ephemeris)
        else:
            self.ephemeris = None
            warnings.warn("WARNING: Planet initiated without Ephemeris")


    def _from_dataframe(self, catalog, koi_id, planet_no):
        df = catalog.loc[catalog.koi_id == koi_id].sort_values(by='period').reset_index(drop=True)

        self.koi_id = koi_id
        self.kic_id = str(df.at[planet_no, 'kic_id'])
        self.period = float(df.at[planet_no, 'period'])
        self.epoch = float(df.at[planet_no, 'epoch'])
        self.depth = float(df.at[planet_no, 'depth']) * 1e-6       # ppm
        self.duration = float(df.at[planet_no, 'duration']) / 24.  # hrs --> days
        self.impact = float(df.at[planet_no, 'impact'])

        return self
    

    def update_ephemeris(self, ephemeris):
        """
        Update ephemeris and corresponding attributes (period & epoch)

        Args:
          ephemeris (Ephemeris)
        
        Returns:
          Planet : self
        """
        if not np.isclose(self.period, ephemeris.period, rtol=0.1):
            raise ValueError(f"New period ({ephemeris.period:.6f}) differs from old period ({self.period:.6f}) by more than 10%")

        self.ephemeris = ephemeris.update_period_and_epoch()
        self.period = self.ephemeris.period
        self.epoch = self.ephemeris.epoch

        return self


__all__ = ["LiteCurve"]


import astropy
from astropy.io import fits as pyfits
import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as sig

from .constants import *


class LiteCurve:
    def __init__(
        self,
        time=None,
        flux=None,
        error=None,
        cadno=None,
        quarter=None,
        season=None,
        quality=None,
        mask=None,
    ):
        self.time = time
        self.flux = flux
        self.error = error
        self.cadno = cadno
        self.quarter = quarter
        self.season = season
        self.quality = quality
        self.mask = mask

    def clip_outliers(self, kernel_size, sigma_upper, sigma_lower, mask=None):
        """
        Sigma-clip outliers using astropy.stats.sigma_clip() and a median filtered trend
        Applies an additional iteration wrapper to allow for masked cadences

        Parameters
        ----------
            kernel_size : int
                size of the median filter kernel
            sigma_upper : float
                upper sigmga clipping threshold
            sigma_lower : float
                lower sigma clipping threshold
            mask : array-like, bool (optional)
                do not reject cadences within masked regions; useful for protecting transits
        """
        if mask is None:
            mask = np.zeros(len(self.time), dtype="bool")

        loop = True
        count = 0

        while loop:
            # first pass: try median filter
            if loop:
                smoothed = sig.medfilt(self.flux, kernel_size=kernel_size)

                bad = astropy.stats.sigma_clip(
                    self.flux - smoothed,
                    sigma_upper=sigma_upper,
                    sigma_lower=sigma_lower,
                    stdfunc=astropy.stats.mad_std,
                ).mask
                bad = bad * ~mask

            # second pass: try savgol filter (if over 1% of points were flagged with median filter)
            if np.sum(bad) / len(bad) > 0.01:
                smoothed = sig.savgol_filter(
                    self.flux, window_length=kernel_size, polyorder=2
                )

                bad = astropy.stats.sigma_clip(
                    self.flux - smoothed,
                    sigma_upper=sigma_upper,
                    sigma_lower=sigma_lower,
                    stdfunc=astropy.stats.mad_std,
                ).mask
                bad = bad * ~mask

            # third pass: skip outlier rejection
            if np.sum(bad) / len(bad) > 0.01:
                bad = np.zeros_like(mask)

            for k in self.__dict__.keys():
                if type(self.__dict__[k]) is np.ndarray:
                    self.__setattr__(k, self.__dict__[k][~bad])

            mask = mask[~bad]

            if np.sum(bad) == 0:
                loop = False
            else:
                count += 1

            if count >= 3:
                loop = False

        return self

    def remove_flagged_cadences(self, qmask):
        """
        Remove user-specified cadences using a pre-generated mask

        Parameters
        ----------
        qmask : array-like
            boolean array; 1=good, 0=bad
        """
        for k in self.__dict__.keys():
            if type(self.__dict__[k]) is np.ndarray:
                self.__setattr__(k, self.__dict__[k][qmask])

        return self

    def plot(self):
        """
        Plot the photometry
        """
        fig = plt.figure(figsize=(20, 4))
        plt.plot(self.time, self.flux, "k", lw=0.5)
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)
        plt.xlabel("Time [BKJD]", fontsize=24)
        plt.ylabel("Flux", fontsize=24)
        plt.xlim(self.time.min(), self.time.max())

        return fig

    def to_fits(self, target, filename, cadence):
        """
        Save LiteCurve object as a fits file

        Parameters
        ----------
        target : string
            name of target
        filename : string
            where to save the fits file
        """
        # make primary HDU
        primary_hdu = pyfits.PrimaryHDU()

        header = primary_hdu.header

        header["TARGET"] = target
        header["CADENCE"] = cadence

        primary_hdu.header = header

        # add it to HDU list
        hdulist = []
        hdulist.append(primary_hdu)

        hdulist.append(pyfits.ImageHDU(self.time, name="TIME"))
        hdulist.append(pyfits.ImageHDU(self.flux, name="FLUX"))
        hdulist.append(pyfits.ImageHDU(self.error, name="ERROR"))
        hdulist.append(pyfits.ImageHDU(self.cadno, name="CADNO"))
        hdulist.append(pyfits.ImageHDU(self.quarter, name="QUARTER"))

        hdulist = pyfits.HDUList(hdulist)
        hdulist.writeto(filename, overwrite=True)

        return None




import os

import requests
from bs4 import BeautifulSoup
import logging
import pandas as pd
import re
import subprocess
import tempfile
import urllib.request
import datetime


class KeplerCatalog:
    def __init__(self, project_root, catalog_dir): 
        self.project_root = project_root
        self.catalog_dir = catalog_dir
        self.catalog = pd.read_csv(self.catalog_dir, index_col=0)

    def get_kic_ids(self, koi_ids):
        if isinstance(koi_ids, str):
            koi_ids = [koi_ids]
        kic_ids = []
        for koi in koi_ids:
            match = self.catalog[self.catalog.koi_id == koi]
            if not match.empty:
                kic_ids.append(str(match['kic_id'].values[0]))
        return kic_ids

class KeplerDataFetcher:
    def __init__(self, kic_id, get_kepler_data_py_path, output_dir, log_file='kepler_fetcher.log'):
        self.kic_id = kic_id
        self.script_path = get_kepler_data_py_path
        self.output_dir = output_dir
        os.makedirs(output_dir, exist_ok=True)

        logging.basicConfig(
            handlers=[
                logging.FileHandler(log_file),
                logging.StreamHandler()
            ],
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )

        logging.info("=" * 60)
        logging.info(f"Starting processing for KIC {self.kic_id}")

    def get_parent_url(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            command = [
                "python", self.script_path,
                str(self.kic_id),
                "-c", "long",
                "--cmdtype", "wget"
            ]
            subprocess.run(command, cwd=tmpdir, check=True)
            logging.info("Ran get_kepler_data.py successfully")

            sh_path = os.path.join(tmpdir, "get_kepler_data.sh")
            with open(sh_path, "r") as f:
                for line in f:
                    match = re.search(r"(http.*/)(kplr\d+-\d+_llc\.fits)", line)
                    if match:
                        return match.group(1)  # parent directory URL
        raise RuntimeError("Could not find a valid parent directory URL.")

    def download_llc_fits_from_directory(self, base_url):

        response = requests.get(base_url)
        response.raise_for_status()
        soup = BeautifulSoup(response.text, "html.parser")

        fits_links = [
            a['href'] for a in soup.find_all('a', href=True)
            if a['href'].endswith("_llc.fits") or a['href'].endswith("_slc.fits")
        ]

        missing = []
        for link in fits_links:
            fits_url = base_url + link
            local_path = os.path.join(self.output_dir, link)
            if os.path.exists(local_path):
                logging.info(f"Already exists: {link}")
                continue

            logging.info(f"Downloading: {fits_url}")
            try:
                r = requests.get(fits_url)
                r.raise_for_status()
                with open(local_path, "wb") as f:
                    f.write(r.content)
            except Exception as e:
                logging.error(f"Failed to download {fits_url}: {e}")
                missing.append(fits_url)
        return missing

    def run(self, max_retries=2):
        base_url = self.get_parent_url()
        for attempt in range(max_retries):
            missing = self.download_llc_fits_from_directory(base_url)
            if not missing:
                logging.info(f"All _llc.fits and _slc.fits files downloaded for KIC {self.kic_id}.")
                return True
            else:
                logging.warning(f"Retrying {len(missing)} files... (Attempt {attempt + 2})")
        logging.error(f"Final missing files for KIC {self.kic_id}: {missing}")
        return missing

class KeplerDownloader:
    @staticmethod
    def run_all(koi_ids, catalog_path, script_path, output_dir, project_root="."):
        catalog = KeplerCatalog(project_root=project_root, catalog_dir=catalog_path)
        kic_ids = catalog.get_kic_ids(koi_ids)

        if not kic_ids:
            print("No matching KIC_IDs found for the given KOI_IDs.")
            return

        for koi_id, kic_id in zip(koi_ids, kic_ids):
            print(f"Processing KOI_ID: {koi_id} → KIC_ID: {kic_id}")

            # Generate a unique timestamp for the log filename
            timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            safe_koi_id = koi_id.replace(" ", "_").replace("/", "_")
            log_filename = f"kepler_fetcher_koi_id_{safe_koi_id}_{timestamp}.log"
            log_file_path = os.path.join(output_dir, log_filename)

            fetcher = KeplerDataFetcher(
                kic_id=kic_id,
                get_kepler_data_py_path=script_path,
                output_dir=output_dir,
                log_file=log_file_path
            )

            result = fetcher.run(max_retries=2)

            if result is True:
                print(f"All files for {koi_id} downloaded successfully.\n")
            else:
                print(f"Some files for {koi_id} could not be downloaded. See log for details.\n")



__all__ = ["uniform_ppf", "loguniform_ppf", "norm_ppf", "prior_transform", "lnlike"]


from batman import _rsky
from batman import _quadratic_ld
from celerite2 import GaussianProcess
import numpy as np
from scipy.special import erfinv

from .constants import *


def uniform_ppf(u, a, b):
    """
    Transform from U(0,1) --> Uniform(a,b)

    Parameters
    ----------
        u : array-like
            samples from standard uniform distribution U(0,1)
        a : float
            lower bound on transformed distribution U(a,b)
        b : float
            upper bound on transformed distribution U(a,b)

    Returns
    -------
        u_t : array-like
            transformed samples from uniform distribution
    """
    return u * (b - a) + a


def loguniform_ppf(u, a, b):
    """
    Transform from U(0,1) --> LogUniform(a,b)

    Parameters
    ----------
        u : array-like
            samples from standard uniform distribution U(0,1)
        a : float
            lower bound on transformed distribution lnU(a,b)
        b : float
            upper bound on transformed distribution lnU(a,b)

    Returns
    -------
        u_t : array-like
            transformed samples from log-uniform distribution
    """
    return np.exp(u * np.log(b) + (1 - u) * np.log(a))


def norm_ppf(u, mu, sig, eps=1e-12):
    """
    Transform from U(0,1) --> Normal(mu,sig)

    Parameters
    ----------
        u : array-like
            samples from standard uniform distribution U(0,1)
        mu : float
            mean of transformed distribution N(mu,sig)
        sig : float
            standard deviation of transformed distribution N(mu,sig)
        eps : float (optional)
            provides numerical stability in the inverse error function at edges

    Returns
    -------
        u_t : array-like
            transformed samples from Normal distribution
    """
    return mu + sig * np.sqrt(2) * erfinv((2 * u - 1) / (1 + eps))


def prior_transform(uniform_hypercube, num_planets, durations):
    """
    Prior transform over physical ALDERAAN basis {C0, C1, r, b, T14}...{q1,q2}

    Distributions are hard-coded to be:
        Normal on ephemeris perturbations {C0,C1}
        Log-uniform on radius ratio r and transit duration T14
        Uniform on impact parameter b
        Uniform on quadratic limb darkening coefficients {q1,q2}

    For motivation behind this prior-parameter choice see:
        Carter+ 2008 (2008ApJ...689..499C)
        Kipping 2013 (2013MNRAS.435.2152K)
        Gilbert, MacDougall, & Petigura 2022 (2022AJ....164...92G)
        MacDougal, Gilbert, & Petigura 2023 (2023AJ....166...61M)

    Parameters
    ----------
        uniform_hypercube : array-like, length 5xN + 2
            list of parameters N x {C0, C1, r, b, T14} + {q1,q2}
        num_planets : int
            number of planets in system
        durations : array-like
            list of transit durations for N planets (used for setting prior limits)

    Returns
    -------
        transformed_hypercube : array_like, length 5xN + 2
            transformed samples
    """
    if num_planets != len(durations):
        raise ValueError("input durations must match provided num_planets")

    u_ = np.array(uniform_hypercube)
    x_ = np.zeros_like(u_)

    # 5*num_planets (+2) parameters: {C0, C1, r, b, T14}...{q1,q2}
    for npl in range(num_planets):
        x_[5 * npl + 0] = norm_ppf(u_[0 + npl * 5], 0.0, 0.1)
        x_[5 * npl + 1] = norm_ppf(u_[1 + npl * 5], 0.0, 0.1)
        x_[5 * npl + 2] = loguniform_ppf(u_[2 + npl * 5], 1e-5, 0.99)
        x_[5 * npl + 3] = uniform_ppf(u_[3 + npl * 5], 0.0, 1 + x_[5 * npl + 2])
        x_[5 * npl + 4] = loguniform_ppf(u_[4 + npl * 5], scit, 3 * durations[npl])

    # limb darkening coefficients (see Kipping 2013)
    x_[-2] = uniform_ppf(u_[-2], 0, 1)
    x_[-1] = uniform_ppf(u_[-1], 0, 1)

    return x_


def lnlike(x, num_planets, theta, ephem_args, phot_args, ld_priors, gp_kernel=None):
    """
    Log-likelihood function to be passed to dynesty sampler

    *** Much of this function is redundant and can be cleaned up ***

    Parameters
    ----------
        x : array-like
            N x [C0, C1, r, b, T14] + [q1,q2]
        num_planets : int
            number of planets in the system
        theta : batman.TransitParams() object

        ephem_args : dict
            arguments related to the ephemeris
        phot_args : dict
            arguments related to the photometry
        ld_priors : tuple
            precomputed (U1,U2) values for limb darkening; used to set priors
        gp_kernel : celerite kernel
            pre-initialized celerite Gaussian Process kernel (optional)

    Returns
    -------
        loglike : float
            log-likelihood

    """
    # extract ephemeris kwargs
    inds = ephem_args["transit_inds"]
    tts = ephem_args["transit_times"]
    legx = ephem_args["transit_legx"]

    # calculate physical limb darkening (see Kipping 2013)
    q1, q2 = np.array(x[-2:])
    u1 = 2 * np.sqrt(q1) * q2
    u2 = np.sqrt(q1) * (1 - 2 * q2)

    # set planet paramters
    for npl in range(num_planets):
        C0, C1, rp, b, T14 = np.array(x[5 * npl : 5 * (npl + 1)])

        theta[npl].rp = rp
        theta[npl].b = b
        theta[npl].T14 = T14
        theta[npl].u = [u1, u2]
        theta[npl].limb_dark = "quadratic"

    # calculate likelihood
    loglike = 0.0

    for j, q in enumerate(phot_args["quarters"]):
        f_ = phot_args["flux"][j]
        e_ = phot_args["error"][j]
        light_curve = np.ones(len(f_), dtype="float")

        for npl in range(num_planets):
            t_ = phot_args["warped_t"][npl][j]
            x_ = phot_args["warped_x"][npl][j]
            C0 = x[5 * npl]
            C1 = x[5 * npl + 1]

            t_ = t_ + C0 + C1 * x_

            exp_time = phot_args["exptime"][q]
            texp_offsets = phot_args["texp_offsets"][q]
            supersample_factor = phot_args["oversample"][q]
            t_supersample = (texp_offsets + t_.reshape(t_.size, 1)).flatten()

            nthreads = 1
            ds = _rsky._rsky(
                t_supersample,
                theta[npl].t0,
                theta[npl].per,
                theta[npl].rp,
                theta[npl].b,
                theta[npl].T14,
                1,
                nthreads,
            )

            # look into the transit type argument
            qld_flux = _quadratic_ld._quadratic_ld(
                ds, np.abs(theta[npl].rp), theta[npl].u[0], theta[npl].u[1], nthreads
            )
            qld_flux = np.mean(
                qld_flux.reshape(-1, supersample_factor), axis=1
            )  # PERF can probably speed this up.
            light_curve += qld_flux - 1.0

        USE_GP = False
        if USE_GP:
            gp = GaussianProcess(gp_kernel[q % 4], mean=light_curve)
            gp.compute(t_, yerr=e_)
            loglike += gp.log_likelihood(f_)
        else:
            loglike += -0.5 * np.sum(((light_curve - f_) / e_) ** 2)

    # enforce prior on limb darkening
    U1, U2 = ld_priors
    sig_ld_sq = 0.01
    loglike -= 1.0 / (2 * sig_ld_sq) * (u1 - U1) ** 2
    loglike -= 1.0 / (2 * sig_ld_sq) * (u2 - U2) ** 2

    if not np.isfinite(loglike):
        return -1e300

    return loglike



#!/usr/bin/env python
# coding: utf-8

# # Detrend and Estimate TTVs


import os
import sys
import glob
import shutil
import warnings
import argparse
from copy import deepcopy
from datetime import datetime
from timeit import default_timer as timer

import astropy
from astropy.stats import mad_std
from astropy.timeseries import LombScargle
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.polynomial.polynomial as poly
from scipy import stats
from scipy.ndimage import median_filter

import pymc3 as pm
import pymc3_ext as pmx
import exoplanet as exo
import aesara_theano_fallback.tensor as T
from aesara_theano_fallback import aesara as theano


# #### Flush buffer and silence extraneous warnings


# flush buffer to avoid mixed outputs from progressbar
sys.stdout.flush()

# turn off FutureWarnings
warnings.filterwarnings("ignore", category=FutureWarning)

# supress UnitsWarnings (this code doesn't use astropy units)
warnings.filterwarnings(
    action="ignore", category=astropy.units.UnitsWarning, module="astropy"
)


# #### Initialize timer


print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Detrending and TTV Estimation")
print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
print("+" * shutil.get_terminal_size().columns)
print("")

# start program timer
global_start_time = timer()


# #### Parse inputs


try:
    parser = argparse.ArgumentParser(
        description="Inputs for ALDERAAN transit fiting pipeline"
    )
    parser.add_argument(
        "--mission",
        default=None,
        type=str,
        required=True,
        help="Mission name; can be 'Kepler' or 'Kepler-Validation'",
    )
    parser.add_argument(
        "--target",
        default=None,
        type=str,
        required=True,
        help="Target name; format should be K00000",
    )
    parser.add_argument(
        "--project_dir",
        default=None,
        type=str,
        required=True,
        help="Project directory for saving outputs",
    )
    parser.add_argument(
        "--data_dir",
        default=None,
        type=str,
        required=True,
        help="Data directory for accessing MAST lightcurves",
    )
    parser.add_argument(
        "--catalog",
        default=None,
        type=str,
        required=True,
        help="CSV file containing input planetary parameters",
    )
    parser.add_argument(
        "--run_id", default=None, type=str, required=True, help="run identifier"
    )
    parser.add_argument(
        "--verbose",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable verbose logging",
    )
    parser.add_argument(
        "--iplot",
        default=False,
        type=bool,
        required=False,
        help="'True' to enable interactive matplotlib backend; default 'agg'",
    )
    parser.add_argument(
        "--use_sc",
        default=False,
        type=bool,
        required=False,
        help="'True' to use short cadence data where available",
    )

    args = parser.parse_args()
    MISSION = args.mission
    TARGET = args.target
    PROJECT_DIR = args.project_dir
    DATA_DIR = args.data_dir
    CATALOG = args.catalog
    RUN_ID = args.run_id
    VERBOSE = args.verbose
    IPLOT = args.iplot
    USE_SC = args.use_sc

except SystemExit:
    warnings.warn("No arguments were parsed from the command line")


print("")
print(f"   MISSION : {MISSION}")
print(f"   TARGET  : {TARGET}")
print(f"   RUN ID  : {RUN_ID}")
print("")
print(f"   Project directory : {PROJECT_DIR}")
print(f"   Data directory    : {DATA_DIR}")
print(f"   Input catalog     : {CATALOG}")
print("")
print(f"   theano cache : {theano.config.compiledir}")
print("")


# #### Build directory structure


# directories in which to place pipeline outputs for this run
RESULTS_DIR = os.path.join(PROJECT_DIR, "Results", RUN_ID, TARGET)
FIGURE_DIR = os.path.join(PROJECT_DIR, "Figures", RUN_ID, TARGET)

os.makedirs(RESULTS_DIR, exist_ok=True)
os.makedirs(FIGURE_DIR, exist_ok=True)

sys.path.append(PROJECT_DIR)


# #### Import ALDERAAN routines


import alderaan.io as io
import alderaan.detrend as detrend
import alderaan.omc as omc
from alderaan.astro import (
    get_transit_depth,
    predict_tc_error,
    make_transit_mask,
    set_oversample_factor,
)
from alderaan.constants import lcit, scit
from alderaan.plotting import (
    plot_holczer,
    plot_tc_vs_chisq,
    plot_omc,
    plot_omc_model_selection,
    plot_folded_transit,
)
from alderaan.utils import boxcar_smooth, bin_data, LS_estimator
from alderaan.validate import remove_known_transits, inject_synthetic_transits
from alderaan.Planet import Planet


# #### Set matplotlib backends


if not IPLOT:
    mpl.use("agg")

if np.any(np.array(["agg", "png", "svg", "pdf", "ps"]) == mpl.get_backend()):
    IPLOT = False


# MAIN SCRIPT BEGINS HERE
def main():
    # # ################
    # # --- DATA I/O ---
    # # ################

    print("\nLoading data...\n")

    # ## Read in planet and star properties

    catalog = io.parse_catalog(
        os.path.join(PROJECT_DIR, f"Catalogs/{CATALOG}"), MISSION, TARGET
    )

    KOI_ID = catalog.koi_id.to_numpy()[0]
    KIC_ID = catalog.kic_id.to_numpy()[0]

    NPL = catalog.npl.to_numpy()[0]

    PERIODS = catalog.period.to_numpy()
    EPOCHS = catalog.epoch.to_numpy()
    DEPTHS = catalog.depth.to_numpy() * 1e-6
    DURS = catalog.duration.to_numpy() / 24
    IMPACTS = catalog.impact.to_numpy()

    U1 = catalog.limbdark_1.to_numpy()[0]
    U2 = catalog.limbdark_2.to_numpy()[0]

    if np.any(np.diff(PERIODS) <= 0):
        raise ValueError("Planets should be ordered by ascending period")

    # ## Read in pre-downloaded lightcurve data

    mast_files = glob.glob(DATA_DIR + f"kplr{KIC_ID:09d}*.fits")
    mast_files.sort()

    # short cadence data
    if USE_SC:
        sc_raw_data = io.read_mast_files(mast_files, KIC_ID, "short cadence")
        sc_quarters = sc_raw_data.quarter
    else:
        sc_raw_data = []
        sc_quarters = []

    sc_data = []
    for i, scrd in enumerate(sc_raw_data):
        sc_data.append(io.LightKurve_to_LiteCurve(scrd))

    # long cadence data
    lc_raw_data = io.read_mast_files(
        mast_files, KIC_ID, "long cadence", exclude=sc_quarters
    )

    lc_data = []
    for i, lcrd in enumerate(lc_raw_data):
        lc_data.append(io.LightKurve_to_LiteCurve(lcrd))

    # # ####################
    # # --- PRELIMINARIES ---
    # # ####################

    # ## Establish time baseline

    print("Establishing observation baseline")

    time_min = []
    time_max = []

    for i, scd in enumerate(sc_data):
        time_min.append(scd.time.min())
        time_max.append(scd.time.max())

    for i, lcd in enumerate(lc_data):
        time_min.append(lcd.time.min())
        time_max.append(lcd.time.max())

    TIME_START = np.min(time_min)
    TIME_END = np.max(time_max)

    if TIME_START < 0:
        raise ValueError("START TIME is negative...this will cause problems")

    # put epochs in range (TIME_START, TIME_START + PERIOD)
    for n in range(NPL):
        if EPOCHS[n] < TIME_START:
            adj = 1 + (TIME_START - EPOCHS[n]) // PERIODS[n]
            EPOCHS[n] += adj * PERIODS[n]

        if EPOCHS[n] > (TIME_START + PERIODS[n]):
            adj = (EPOCHS[n] - TIME_START) // PERIODS[n]
            EPOCHS[n] -= adj * PERIODS[n]

    # ## Initialize Planet objects

    print(f"Initializing {NPL} Planet objects")

    planets = []
    for n in range(NPL):
        p = Planet()

        # put in some basic transit parameters
        p.period = PERIODS[n]
        p.epoch = EPOCHS[n]
        p.depth = DEPTHS[n]
        p.duration = DURS[n]
        p.impact = IMPACTS[n]

        if p.impact > 1 - np.sqrt(p.depth):
            p.impact = (1 - np.sqrt(p.depth)) ** 2

        # estimate transit times from linear ephemeris
        p.tts = np.arange(p.epoch, TIME_END, p.period)

        # make transit indexes
        p.index = np.array(np.round((p.tts - p.epoch) / p.period), dtype="int")

        # add to list
        planets.append(p)

    # ## Set oversample factor

    oversample_lc = set_oversample_factor(
        PERIODS, DEPTHS, DURS, lc_data[0].flux, lc_data[0].error, lcit
    )
    oversample_sc = 1

    if oversample_lc >= 60:
        raise ValueError("attempting to set oversample factor greater than 60")

    print(f"Setting oversample = {oversample_lc}")

    # ## Build initial TTV model

    print("\nBuilding initial TTV model...\n")

    # load, clean and match Holczer+2016 data
    h16file = os.path.join(PROJECT_DIR, "Catalogs/holczer_2016_kepler_ttvs.txt")
    holczer = omc.load_holczer_ttvs(h16file, NPL, KOI_ID)
    holczer = omc.clean_holczer_ttvs(holczer, TIME_START, TIME_END)
    planets = omc.match_holczer_ttvs(planets, holczer)

    # make some plots
    for n in range(NPL):
        if np.isfinite(holczer["period"][n]):
            fig = plot_holczer(holczer, n)

            figpath = os.path.join(
                FIGURE_DIR, TARGET + f"_ttvs_holczer_{holczer['period'][n]:.1f}.png"
            )
            plt.savefig(figpath, bbox_inches="tight")
            if IPLOT:
                plt.show()
            else:
                plt.close()

    # # ##################################
    # # ----- INJECTION-AND-RECOVERY -----
    # # ##################################

    if MISSION == "Kepler-Validation":
        # remove known transits of real planets
        if len(lc_data) > 0:
            lc_data = remove_known_transits(
                planets, lc_data, [U1, U2], lcit, oversample_lc
            )

        if len(sc_data) > 0:
            sc_data = remove_known_transits(
                planets, sc_data, [U1, U2], scit, oversample_sc
            )

        # load synthetic catalog
        path = os.path.join(PROJECT_DIR, f"Simulations/{RUN_ID}/{RUN_ID}.csv")
        catalog = io.parse_catalog(path, MISSION, TARGET)

        KOI_ID = catalog.koi_id.to_numpy()[0]
        KIC_ID = catalog.kic_id.to_numpy()[0]

        NPL = catalog.npl.to_numpy()[0]

        PERIODS = catalog.period.to_numpy()
        EPOCHS = catalog.epoch.to_numpy()
        DEPTHS = catalog.depth.to_numpy() * 1e-6
        DURS = catalog.duration.to_numpy() / 24
        IMPACTS = catalog.impact.to_numpy()

        U1 = catalog.limbdark_1.to_numpy()[0]
        U2 = catalog.limbdark_2.to_numpy()[0]

        if np.any(np.diff(PERIODS) <= 0):
            raise ValueError("Planets should be ordered by ascending period")

        # put epochs in range (TIME_START, TIME_START + PERIOD)
        for n in range(NPL):
            if EPOCHS[n] < TIME_START:
                adj = 1 + (TIME_START - EPOCHS[n]) // PERIODS[n]
                EPOCHS[n] += adj * PERIODS[n]

            if EPOCHS[n] > (TIME_START + PERIODS[n]):
                adj = (EPOCHS[n] - TIME_START) // PERIODS[n]
                EPOCHS[n] -= adj * PERIODS[n]

        # initalize synthetic planet objects
        planets = []
        for n in range(NPL):
            p = Planet()

            # put in some basic transit parameters
            p.period = PERIODS[n]
            p.epoch = EPOCHS[n]
            p.depth = DEPTHS[n] * 1e6
            p.duration = DURS[n] * 24
            p.impact = IMPACTS[n]

            # load true transit times
            tts_file = os.path.join(
                PROJECT_DIR, f"Simulations/{RUN_ID}/S{TARGET[1:]}_{n}.tts"
            )
            true_tts = np.loadtxt(tts_file).swapaxes(0, 1)

            p.tts = true_tts[1]
            p.index = np.array(true_tts[0], dtype="int")

            planets.append(p)

        # inject synthetic transits
        if len(lc_data) > 0:
            lc_data = inject_synthetic_transits(
                planets, lc_data, [U1, U2], lcit, oversample_lc
            )

        if len(sc_data) > 0:
            sc_data = inject_synthetic_transits(
                planets, sc_data, [U1, U2], scit, oversample_sc
            )

        lc_raw_sim_data = deepcopy(lc_data)
        sc_raw_sim_data = deepcopy(sc_data)

    # # #########################
    # # ----- 1ST DETRENDING -----
    # # #########################

    print("\nDetrending lightcurves (1st pass)...\n")

    # ## Detrend the lightcurves

    # make transit masks
    for i, lcd in enumerate(lc_data):
        mask = np.zeros((NPL, len(lcd.time)), dtype="bool")

        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                lcd.time, p.tts, np.max([1 / 24, 1.5 * p.duration])
            )

        lcd.mask = mask.sum(axis=0) > 0

    for i, scd in enumerate(sc_data):
        mask = np.zeros((NPL, len(scd.time)), dtype="bool")

        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                scd.time, p.tts, np.max([1 / 24, 1.5 * p.duration])
            )

        scd.mask = mask.sum(axis=0) > 0

    # remove bad cadences
    for i, lcd in enumerate(lc_data):
        lcd = detrend.remove_bad_cadences(
            lcd, planets, rel_masksize=1.5, min_masksize=1.0 / 24
        )

    for i, scd in enumerate(sc_data):
        scd = detrend.remove_bad_cadences(
            scd, planets, rel_masksize=1.5, min_masksize=1.0 / 24
        )

    # track stellar and instrumental oscillations
    oscillation_period_by_quarter = np.ones(18) * np.nan

    for i, lcd in enumerate(lc_data):
        min_period = np.max([5 * np.max(DURS), 13 * lcit])
        oscillation_period_by_quarter[
            lcd.quarter[0]
        ] = detrend.estimate_oscillation_period(lcd, min_period=min_period)

    for i, scd in enumerate(sc_data):
        min_period = np.max([5 * np.max(DURS), 91 * scit])
        oscillation_period_by_quarter[
            scd.quarter[0]
        ] = detrend.estimate_oscillation_period(scd, min_period=min_period)

    oscillation_period_by_season = np.zeros((4, 2))

    for i in range(4):
        oscillation_period_by_season[i, 0] = np.nanmedian(
            oscillation_period_by_quarter[i::4]
        )
        oscillation_period_by_season[i, 1] = mad_std(
            oscillation_period_by_quarter[i::4], ignore_nan=True
        )

    # detrend long cadence data
    break_tolerance = np.max([int(DURS.min() / lcit * 5 / 2), 13])
    min_per = np.max([5 * np.max(DURS), 13 * lcit])

    for i, lcd in enumerate(lc_data):
        print(f"QUARTER {lcd.quarter[0]}")

        nom_per = oscillation_period_by_season[lcd.quarter[0] % 4][0]

        lcd = detrend.flatten_litecurve(
            lcd, break_tolerance, min_per, nominal_period=nom_per, verbose=VERBOSE
        )

    if len(lc_data) > 0:
        lc = detrend.stitch(lc_data)
    else:
        lc = None

    if lc is not None:
        fig = lc.plot()

        figpath = os.path.join(FIGURE_DIR, TARGET + f"_lightcurve_detrended_lc.png")
        plt.savefig(figpath, bbox_inches="tight")
        if IPLOT:
            plt.show()
        else:
            plt.close()

    # detrend short cadence data
    break_tolerance = np.max([int(DURS.min() / scit * 5 / 2), 91])
    min_per = np.max([5 * np.max(DURS), 91 * scit])

    for i, scd in enumerate(sc_data):
        print(f"QUARTER {scd.quarter[0]}")

        nom_per = oscillation_period_by_season[scd.quarter[0] % 4][0]

        scd = detrend.flatten_litecurve(
            scd, break_tolerance, min_per, nominal_period=nom_per, verbose=VERBOSE
        )

    if len(sc_data) > 0:
        sc = detrend.stitch(sc_data)
    else:
        sc = None

    if sc is not None:
        fig = sc.plot()

        figpath = os.path.join(FIGURE_DIR, TARGET + f"_lightcurve_detrended_sc.png")
        plt.savefig(figpath, bbox_inches="tight")
        if IPLOT:
            plt.show()
        else:
            plt.close()

    # # ##########################
    # # ----- QUALITY CONTROL -----
    # # ##########################

    # ## Make wide masks that track each planet individually
    # #### These masks have width 2.5 transit durations, which is probably wider than the masks used for detrending

    if sc is not None:
        mask = np.zeros((NPL, len(sc.time)), dtype="bool")
        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                sc.time, p.tts, np.max([2 / 24, 2.5 * p.duration])
            )

        sc.mask = mask.sum(axis=0) > 0

    if lc is not None:
        mask = np.zeros((NPL, len(lc.time)), dtype="bool")
        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                lc.time, p.tts, np.max([2 / 24, 2.5 * p.duration])
            )

        lc.mask = mask.sum(axis=0) > 0

    # ## Flag high quality transits (quality = 1)
    # #### Good transits must have  at least 50% photometry coverage in/near transit

    # number of photometric points expected in transit
    count_expect_lc = np.zeros(NPL, dtype="int")
    count_expect_sc = np.zeros(NPL, dtype="int")

    for n, p in enumerate(planets):
        count_expect_lc[n] = np.max([1, int(np.floor(p.duration / lcit))])
        count_expect_sc[n] = np.max([1, int(np.floor(p.duration / scit))])
        quality = np.zeros(len(p.tts), dtype="bool")

        for i, t0 in enumerate(p.tts):
            if sc is not None:
                in_sc = np.abs(sc.time - t0) / p.duration < 0.5
                near_sc = np.abs(sc.time - t0) / p.duration < 1.5

                qual_in = np.sum(in_sc) > 0.5 * count_expect_sc[n]
                qual_near = np.sum(near_sc) > 1.5 * count_expect_sc[n]

                quality[i] += qual_in * qual_near

            if lc is not None:
                in_lc = np.abs(lc.time - t0) / p.duration < 0.5
                near_lc = np.abs(lc.time - t0) / p.duration < 1.5

                qual_in = np.sum(in_lc) > 0.5 * count_expect_lc[n]
                qual_near = np.sum(near_lc) > 1.5 * count_expect_lc[n]

                quality[i] += qual_in * qual_near

        p.quality = np.copy(quality)

        if np.sum(p.quality) < 0.5 * len(p.quality):
            raise ValueError(
                f"Over 50% of transits for Planet {n} have been flagged as low quality"
            )

    # ## Flag overlapping transits

    dur_max = np.max(DURS)
    overlap = [None] * NPL

    for i in range(NPL):
        overlap[i] = np.zeros(len(planets[i].tts), dtype="bool")

        for j in range(NPL):
            if i != j:
                for ttj in planets[j].tts:
                    overlap[i] += np.abs(planets[i].tts - ttj) / dur_max < 1.5

    # ## Count up transits and calculate initial fixed transit times

    num_transits = np.zeros(NPL, dtype="int")
    transit_inds = []
    fixed_tts = []

    for n, p in enumerate(planets):
        transit_inds.append(np.array((p.index - p.index.min())[p.quality], dtype="int"))
        fixed_tts.append(np.copy(p.tts)[p.quality])

        num_transits[n] = len(transit_inds[n])
        transit_inds[n] -= transit_inds[n].min()

    # ## Grab data near transits

    # go quarter-by-quarter
    all_time = [None] * 18
    all_flux = [None] * 18
    all_error = [None] * 18
    all_dtype = ["none"] * 18

    for q in range(18):
        if sc is not None:
            if np.isin(q, sc.quarter):
                use = (sc.mask) * (sc.quarter == q)

                if np.sum(use) > np.min(count_expect_sc):
                    all_time[q] = sc.time[use]
                    all_flux[q] = sc.flux[use]
                    all_error[q] = sc.error[use]
                    all_dtype[q] = "short"

                else:
                    all_dtype[q] = "short_no_transits"

        if lc is not None:
            if np.isin(q, lc.quarter):
                use = (lc.mask) * (lc.quarter == q)

                if np.sum(use) > np.min(count_expect_lc):
                    all_time[q] = lc.time[use]
                    all_flux[q] = lc.flux[use]
                    all_error[q] = lc.error[use]
                    all_dtype[q] = "long"

                else:
                    all_dtype[q] = "long_no_transits"

    all_dtype = np.array(all_dtype)

    # track which quarters have coverage
    quarters = np.arange(18)[(all_dtype == "short") + (all_dtype == "long")]

    # track mean and variance of each quarter
    mean_by_quarter = np.ones(18) * np.nan
    var_by_quarter = np.ones(18) * np.nan

    for q in range(18):
        if sc is not None:
            if np.isin(q, sc.quarter):
                mean_by_quarter[q] = np.mean(sc.flux[sc.quarter == q])
                var_by_quarter[q] = np.var(sc.flux[sc.quarter == q])

        if lc is not None:
            if np.isin(q, lc.quarter):
                mean_by_quarter[q] = np.mean(lc.flux[lc.quarter == q])
                var_by_quarter[q] = np.var(lc.flux[lc.quarter == q])

    # broadcast oversample factor
    oversample = np.zeros(18, dtype="int")
    oversample[all_dtype == "short"] = oversample_sc
    oversample[all_dtype == "long"] = oversample_lc

    # broadcast exposure times
    texp = np.zeros(18)
    texp[all_dtype == "short"] = scit
    texp[all_dtype == "long"] = lcit

    # ## Pull basic transit parameters

    periods = np.zeros(NPL)
    epochs = np.zeros(NPL)
    depths = np.zeros(NPL)
    durs = np.zeros(NPL)
    impacts = np.zeros(NPL)

    for n, p in enumerate(planets):
        periods[n] = p.period
        epochs[n] = p.epoch
        depths[n] = p.depth
        durs[n] = p.duration
        impacts[n] = p.impact

    # ## Define Legendre polynomials

    # Legendre polynomials for better orthogonality; "x" is in the range (-1,1)
    t = [None] * NPL
    x = [None] * NPL
    Leg0 = [None] * NPL
    Leg1 = [None] * NPL

    # this assumes a baseline in the range (TIME_START,TIME_END)
    for n, p in enumerate(planets):
        t[n] = p.epoch + transit_inds[n] * p.period
        x[n] = 2 * (t[n] - TIME_START) / (TIME_END - TIME_START) - 1

        Leg0[n] = np.ones_like(x[n])
        Leg1[n] = x[n].copy()

    print("")
    print("cumulative runtime = ", int(timer() - global_start_time), "s")
    print("")

    # # ############################
    # # ----- LIGHTCURVE FITTING -----
    # # ############################

    # ## Fit transit SHAPE model

    print("\nFitting transit SHAPE model...\n")

    with pm.Model() as shape_model:
        # planetary parameters
        log_r = pm.Uniform(
            "log_r",
            lower=np.log(1e-5),
            upper=np.log(0.99),
            shape=NPL,
            testval=np.log(np.sqrt(depths)),
        )
        r = pm.Deterministic("r", T.exp(log_r))
        b = pm.Uniform("b", lower=0.0, upper=1.0, shape=NPL)

        log_dur = pm.Normal("log_dur", mu=np.log(durs), sd=5.0, shape=NPL)
        dur = pm.Deterministic("dur", T.exp(log_dur))

        # polynomial TTV parameters
        C0 = pm.Normal("C0", mu=0.0, sd=durs / 2, shape=NPL)
        C1 = pm.Normal("C1", mu=0.0, sd=durs / 2, shape=NPL)

        transit_times = []
        for n in range(NPL):
            transit_times.append(
                pm.Deterministic(
                    f"tts_{n}", fixed_tts[n] + C0[n] * Leg0[n] + C1[n] * Leg1[n]
                )
            )

        # set up stellar model and planetary orbit
        starrystar = exo.LimbDarkLightCurve([U1, U2])
        orbit = exo.orbits.TTVOrbit(
            transit_times=transit_times,
            transit_inds=transit_inds,
            b=b,
            ror=r,
            duration=dur,
        )

        # track period and epoch
        pm.Deterministic("P", orbit.period)
        pm.Deterministic("T0", orbit.t0)

        # nuissance parameters
        mbq = mean_by_quarter[(all_dtype == "short") + (all_dtype == "long")]
        vbq = var_by_quarter[(all_dtype == "short") + (all_dtype == "long")]

        flux0 = pm.Normal("flux0", mu=mbq, sd=mbq / 10, shape=len(quarters))
        log_jit = pm.Normal("log_jit", mu=np.log(vbq), sd=10, shape=len(quarters))

        # now evaluate the model for each quarter
        light_curves = [None] * len(quarters)
        model_flux = [None] * len(quarters)
        flux_err = [None] * len(quarters)
        obs = [None] * len(quarters)

        for j, q in enumerate(quarters):
            # calculate light curves
            light_curves[j] = starrystar.get_light_curve(
                orbit=orbit, r=r, t=all_time[q], oversample=oversample[q], texp=texp[q]
            )

            model_flux[j] = pm.math.sum(light_curves[j], axis=-1) + flux0[j] * T.ones(
                len(all_time[q])
            )
            flux_err[j] = T.sqrt(
                np.mean(all_error[q]) ** 2 + T.exp(log_jit[j])
            ) / np.sqrt(2)

            obs[j] = pm.Normal(
                f"obs_{j}", mu=model_flux[j], sd=flux_err[j], observed=all_flux[q]
            )

    with shape_model:
        shape_map = shape_model.test_point
        shape_map = pmx.optimize(
            start=shape_map, vars=[flux0, log_jit], progress=VERBOSE
        )
        shape_map = pmx.optimize(start=shape_map, vars=[b, r, dur], progress=VERBOSE)
        shape_map = pmx.optimize(start=shape_map, vars=[C0, C1], progress=VERBOSE)
        shape_map = pmx.optimize(start=shape_map, progress=VERBOSE)

    # grab transit times and ephemeris
    shape_transit_times = []
    shape_linear_ephemeris = []

    for n, p in enumerate(planets):
        shape_transit_times.append(shape_map[f"tts_{n}"])
        shape_linear_ephemeris.append(
            shape_map["P"][n] * transit_inds[n] + shape_map["T0"][n]
        )

    # update parameter values
    periods = np.atleast_1d(shape_map["P"])
    epochs = np.atleast_1d(shape_map["T0"])
    depths = np.atleast_1d(get_transit_depth(shape_map["r"], shape_map["b"]))
    durs = np.atleast_1d(shape_map["dur"])
    impacts = np.atleast_1d(shape_map["b"])
    rors = np.atleast_1d(shape_map["r"])

    for n, p in enumerate(planets):
        p.period = periods[n]
        p.epoch = epochs[n]
        p.depth = depths[n]
        p.duration = durs[n]
        p.impact = impacts[n]
        p.ror = rors[n]

    print("")
    print("cumulative runtime = ", int(timer() - global_start_time), "s")
    print("")

    # ## Fit SLIDE TTVs

    print("\nFitting TTVs..\n")

    # get list of threshold times between quarters
    qthresh = np.zeros(len(quarters) + 1)
    qthresh[0] = TIME_START - 0.5

    for j, q in enumerate(quarters):
        if lc is not None:
            if np.isin(q, np.unique(lc.quarter)):
                qthresh[j + 1] = lc.time[lc.quarter == q].max() + lcit

        if sc is not None:
            if np.isin(q, np.unique(sc.quarter)):
                qthresh[j + 1] = sc.time[sc.quarter == q].max() + scit

    qthresh[-1] = TIME_END + 0.5

    # track which quarter each transit falls in
    transit_quarter = [None] * NPL

    for n in range(NPL):
        tts = shape_transit_times[n]
        transit_quarter[n] = np.zeros(len(tts), dtype="int")

        for j, q in enumerate(quarters):
            transit_quarter[n][(tts >= qthresh[j]) * (tts < qthresh[j + 1])] = q

    slide_transit_times = []
    slide_uncertainty = []

    for n, p in enumerate(planets):
        print("\nPLANET", n)

        slide_transit_times.append([])
        slide_uncertainty.append([])

        # create template transit
        starrystar = exo.LimbDarkLightCurve([U1, U2])
        orbit = exo.orbits.KeplerianOrbit(
            t0=0, period=p.period, b=p.impact, ror=p.ror, duration=p.duration
        )

        slide_offset = 1.0
        template_time = np.arange(0, (0.02 + p.duration) * (slide_offset + 1.6), scit)
        template_time = np.hstack([-template_time[:-1][::-1], template_time])
        template_flux = (
            1.0
            + starrystar.get_light_curve(
                orbit=orbit, r=p.ror, t=template_time, oversample=1
            )
            .sum(axis=-1)
            .eval()
        )

        # empty lists to hold new transit time and uncertainties
        tts = -99 * np.ones_like(shape_transit_times[n])
        err = -99 * np.ones_like(shape_transit_times[n])

        for i, t0 in enumerate(shape_transit_times[n]):
            # print(i, np.round(t0,2))
            if ~overlap[n][p.quality][i]:
                # identify quarter
                q = transit_quarter[n][i]

                # set exposure time and oversample factor
                if all_dtype[q] == "long":
                    exptime = lcit
                    texp_offsets = np.linspace(
                        -exptime / 2.0, exptime / 2.0, oversample[q]
                    )
                elif all_dtype[q] == "short":
                    exptime = scit
                    texp_offsets = np.array([0.0])
                else:
                    raise ValueError("data cadence expected to be 'long' or 'short'")

                # grab data near each non-overlapping transit
                use = np.abs(all_time[q] - t0) / p.duration < 2.5
                mask = np.abs(all_time[q] - t0) / p.duration < 1.0

                t_ = all_time[q][use]
                f_ = all_flux[q][use]
                m_ = mask[use]

                t_supersample = (texp_offsets + t_.reshape(t_.size, 1)).flatten()

                # remove any residual out-of-transit trend
                try:
                    trend = poly.polyval(t_, poly.polyfit(t_[~m_], f_[~m_], 1))
                except TypeError:
                    trend = np.ones_like(f_)

                f_ /= trend
                e_ = np.ones_like(f_) * np.std(f_[~m_])

                # slide along transit time vector and calculate chisq
                gridstep = scit / 1.618 / 3
                tc_vector = np.arange(0, p.duration * slide_offset, gridstep)
                tc_vector = t0 + np.hstack([-tc_vector[:-1][::-1], tc_vector])
                chisq_vector = np.zeros_like(tc_vector)

                for j, tc in enumerate(tc_vector):
                    y_ = np.interp(t_supersample - tc, template_time, template_flux)
                    y_ = bin_data(t_supersample, y_, exptime, bin_centers=t_)[1]

                    chisq_vector[j] = np.sum((f_ - y_) ** 2 / e_**2)

                # grab points near minimum chisq
                delta_chisq = 1.0

                loop = True
                while loop:
                    # incrememnt delta_chisq and find minimum
                    delta_chisq *= 2
                    min_chisq = chisq_vector.min()

                    # grab the points near minimum
                    tcfit = tc_vector[chisq_vector < min_chisq + delta_chisq]
                    x2fit = chisq_vector[chisq_vector < min_chisq + delta_chisq]

                    # eliminate points far from the local minimum
                    spacing = np.median(tcfit[1:] - tcfit[:-1])
                    faraway = (
                        np.abs(tcfit - np.median(tcfit)) / spacing > 1 + len(tcfit) / 2
                    )

                    tcfit = tcfit[~faraway]
                    x2fit = x2fit[~faraway]

                    # check for stopping conditions
                    if len(x2fit) >= 7:
                        loop = False

                    if delta_chisq >= 16:
                        loop = False

                # fit a parabola around the minimum (need at least 3 pts)
                if len(tcfit) < 3:
                    # print("too few points")
                    tts[i] = np.nan
                    err[i] = np.nan

                else:
                    quad_coeffs = np.polyfit(tcfit, x2fit, 2)
                    quad_mod = np.polyval(quad_coeffs, tcfit)
                    qtc_min = -quad_coeffs[1] / (2 * quad_coeffs[0])
                    qx2_min = np.polyval(quad_coeffs, qtc_min)
                    qtc_err = np.sqrt(1 / quad_coeffs[0])

                    # transit time and scaled error
                    tts[i] = np.mean([qtc_min, np.median(tcfit)])
                    err[i] = qtc_err * (1 + np.std(x2fit - quad_mod))

                    # check that the fit is well-conditioned (ie. a negative t**2 coefficient)
                    if quad_coeffs[0] <= 0.0:
                        # print("inverted parabola")
                        tts[i] = np.nan
                        err[i] = np.nan

                    # check that the recovered transit time is within the expected range
                    if (tts[i] < tcfit.min()) or (tts[i] > tcfit.max()):
                        # print("tc out of bounds")
                        tts[i] = np.nan
                        err[i] = np.nan

                # show plots
                do_plots = False
                if do_plots:
                    if ~np.isnan(tts[i]):
                        tc = tts[i]
                    else:
                        tc = t0

                    fig, ax = plot_tc_vs_chisq(
                        t_,
                        f_,
                        template_time,
                        template_flux,
                        tcfit,
                        x2fit,
                        quad_mod,
                        tc,
                        f"C{n}",
                    )

                    if IPLOT:
                        plt.show()
                    else:
                        plt.close()

            else:
                # print("overlapping transits")
                tts[i] = np.nan
                err[i] = np.nan

        slide_transit_times[n] = np.copy(tts)
        slide_uncertainty[n] = np.copy(err)

    # flag transits for which the slide method failed
    for n, p in enumerate(planets):
        bad = np.isnan(slide_transit_times[n]) + np.isnan(slide_uncertainty[n])
        bad += slide_uncertainty[n] > 8 * np.nanmedian(slide_uncertainty[n])

        slide_transit_times[n][bad] = shape_transit_times[n][bad]
        slide_uncertainty[n][bad] = np.nan

    refit = [None] * NPL
    for n in range(NPL):
        refit[n] = np.isnan(slide_uncertainty[n])

        # if every slide fit worked, randomly select a pair of transits for refitting
        # this is easier than tracking the edge cases -- we'll use the slide ttvs in the final vector anyway
        if np.all(~refit[n]):
            refit[n][np.random.randint(len(refit[n]), size=2)] = True

    print("")
    print("cumulative runtime = ", int(timer() - global_start_time), "s")
    print("")

    # ## Fit MAP INDEPENDENT TTVs
    #
    # #### Only refit transit times for which the slide method failed

    if sc is not None:
        refit_mask_sc = np.zeros((NPL, len(sc.time)), dtype="bool")
        for n, p in enumerate(planets):
            tts = slide_transit_times[n][refit[n]]
            refit_mask_sc[n] = make_transit_mask(
                sc.time, tts, np.max([2 / 24, 2.5 * p.duration])
            )

        refit_mask_sc = refit_mask_sc.sum(axis=0) > 0

    else:
        refit_mask_sc = None

    if lc is not None:
        refit_mask_lc = np.zeros((NPL, len(lc.time)), dtype="bool")
        for n, p in enumerate(planets):
            tts = slide_transit_times[n][refit[n]]
            refit_mask_lc[n] = make_transit_mask(
                lc.time, tts, np.max([2 / 24, 2.5 * p.duration])
            )

        refit_mask_lc = refit_mask_lc.sum(axis=0) > 0

    else:
        refit_mask_lc = None

    # go quarter-by-quarter
    refit_time = [None] * 18
    refit_flux = [None] * 18
    refit_error = [None] * 18
    refit_dtype = ["none"] * 18

    for q in range(18):
        if sc is not None:
            if np.isin(q, sc.quarter):
                use = (refit_mask_sc) * (sc.quarter == q)

                if np.sum(use) > np.min(count_expect_sc):
                    refit_time[q] = sc.time[use]
                    refit_flux[q] = sc.flux[use]
                    refit_error[q] = sc.error[use]
                    refit_dtype[q] = "short"

                else:
                    refit_dtype[q] = "short_no_transits"

        if lc is not None:
            if np.isin(q, lc.quarter):
                use = (refit_mask_lc) * (lc.quarter == q)

                if np.sum(use) > np.min(count_expect_lc):
                    refit_time[q] = lc.time[use]
                    refit_flux[q] = lc.flux[use]
                    refit_error[q] = lc.error[use]
                    refit_dtype[q] = "long"

                else:
                    refit_dtype[q] = "long_no_transits"

    refit_dtype = np.array(refit_dtype)

    # track which quarters have coverage
    refit_quarters = np.arange(18)[(refit_dtype == "short") + (refit_dtype == "long")]

    with pm.Model() as indep_model:
        # transit times
        tt_offset = []
        refit_tts = []
        refit_inds = []

        for n in range(NPL):
            use = np.copy(refit[n])

            tt_offset.append(pm.Normal(f"tt_offset_{n}", mu=0, sd=1, shape=np.sum(use)))

            refit_tts.append(
                pm.Deterministic(
                    f"tts_{n}", shape_transit_times[n][use] + tt_offset[n] * durs[n] / 3
                )
            )

            refit_inds.append(transit_inds[n][use])

        # set up stellar model and planetary orbit
        starrystar = exo.LimbDarkLightCurve([U1, U2])
        orbit = exo.orbits.TTVOrbit(
            transit_times=refit_tts,
            transit_inds=refit_inds,
            period=periods,
            b=impacts,
            ror=rors,
            duration=durs,
        )

        # nuissance parameters
        mbq = mean_by_quarter[(refit_dtype == "short") + (refit_dtype == "long")]
        vbq = var_by_quarter[(refit_dtype == "short") + (refit_dtype == "long")]

        flux0 = pm.Normal("flux0", mu=mbq, sd=mbq / 10, shape=len(refit_quarters))
        log_jit = pm.Normal("log_jit", mu=np.log(vbq), sd=10, shape=len(refit_quarters))

        # now evaluate the model for each quarter
        light_curves = [None] * len(refit_quarters)
        model_flux = [None] * len(refit_quarters)
        flux_err = [None] * len(refit_quarters)
        obs = [None] * len(refit_quarters)

        for j, q in enumerate(refit_quarters):
            # calculate light curves
            light_curves[j] = starrystar.get_light_curve(
                orbit=orbit,
                r=rors,
                t=refit_time[q],
                oversample=oversample[q],
                texp=texp[q],
            )

            model_flux[j] = pm.math.sum(light_curves[j], axis=-1) + flux0[j] * T.ones(
                len(refit_time[q])
            )
            flux_err[j] = T.sqrt(
                np.mean(refit_error[q]) ** 2 + T.exp(log_jit[j])
            ) / np.sqrt(2)

            obs[j] = pm.Normal(
                f"obs_{j}", mu=model_flux[j], sd=flux_err[j], observed=refit_flux[q]
            )

    with indep_model:
        indep_map = indep_model.test_point
        indep_map = pmx.optimize(
            start=indep_map, vars=[flux0, log_jit], progress=VERBOSE
        )

        for n in range(NPL):
            indep_map = pmx.optimize(
                start=indep_map, vars=[tt_offset[n]], progress=VERBOSE
            )

        indep_map = pmx.optimize(start=indep_map, progress=VERBOSE)

    indep_transit_times = []
    indep_uncertainty = []
    indep_linear_ephemeris = []
    full_indep_linear_ephemeris = []

    for n, p in enumerate(planets):
        indep_transit_times.append(np.copy(slide_transit_times[n]))
        indep_uncertainty.append(np.copy(slide_uncertainty[n]))

        replace = np.isnan(slide_uncertainty[n])

        if np.any(replace):
            indep_transit_times[n][replace] = indep_map[f"tts_{n}"]

        pfit = poly.polyfit(transit_inds[n], indep_transit_times[n], 1)

        indep_linear_ephemeris.append(poly.polyval(transit_inds[n], pfit))
        full_indep_linear_ephemeris.append(poly.polyval(p.index, pfit))

        if np.all(replace):
            indep_uncertainty[n][replace] = mad_std(
                indep_transit_times[n] - indep_linear_ephemeris[n]
            )
        elif np.any(replace):
            indep_uncertainty[n][replace] = mad_std(
                indep_transit_times[n][~replace] - indep_linear_ephemeris[n][~replace]
            )

    for n, p in enumerate(planets):
        print(f"\nPLANET {n}")

        xtime = indep_linear_ephemeris[n]
        yomc = indep_transit_times[n] - indep_linear_ephemeris[n]

        yerr_expected = predict_tc_error(
            np.sqrt(p.depth), p.impact, p.duration, lcit, np.median(lc.error)
        )
        yerr_measured = indep_uncertainty[n]

        yerr = np.sqrt(yerr_expected**2 + yerr_measured**2)

        print("  expected uncertainty: {0:.1f} min".format(yerr_expected * 24 * 60))
        print(
            "  measured uncertainty: {0:.1f} min".format(
                np.median(yerr_measured) * 24 * 60
            )
        )
        print("  adopted uncertainty:  {0:.1f} min".format(np.median(yerr) * 24 * 60))
        print("  dispersion:           {0:.1f} min".format(mad_std(yomc) * 24 * 60))

    print("")
    print("cumulative runtime = ", int(timer() - global_start_time), "s")
    print("")

    # # ###############################
    # # ----- OMC MODEL SELECTION -----
    # # ###############################

    print("\nIdentifying best OMC model...\n")

    # ## Search for periodic signals

    print("...searching for periodic signals")

    indep_freqs = []
    indep_faps = []

    for n, p in enumerate(planets):
        # grab data
        xtime = indep_linear_ephemeris[n]
        yomc = indep_transit_times[n] - indep_linear_ephemeris[n]

        # flag outliers
        ymed = boxcar_smooth(median_filter(yomc, size=5, mode="mirror"), winsize=5)
        out = np.abs(yomc - ymed) / mad_std(yomc - ymed) > 5.0

        # search for a periodic component
        peakfreq = np.nan
        peakfap = 1.0

        if NPL == 1:
            fap = 0.1
        elif NPL > 1:
            fap = 0.99

        if np.sum(~out) > 8:
            try:
                xf, yf, freqs, faps = LS_estimator(xtime[~out], yomc[~out], fap=fap)

                if len(freqs) > 0:
                    if freqs[0] > xf.min():
                        peakfreq = freqs[0]
                        peakfap = faps[0]

            except Exception:
                pass

        indep_freqs.append(peakfreq)
        indep_faps.append(peakfap)

    omc_freqs = []
    omc_faps = []

    # for single planet systems, use the direct LS output
    if NPL == 1:
        if np.isnan(indep_freqs[0]):
            omc_freqs.append(None)
            omc_faps.append(None)
        else:
            omc_freqs.append(indep_freqs[0])
            omc_faps.append(indep_faps[0])

    # for multiplanet systems, check if any statistically marginal frequencies match between planets
    elif NPL > 1:
        for i in range(NPL):
            # save any low FAP frequencies
            if indep_faps[i] < 0.1:
                omc_freqs.append(indep_freqs[i])
                omc_faps.append(indep_faps[i])

            # check if the LS frequency is close to that of any other planet
            else:
                close = False

                df_min = 1 / (
                    indep_linear_ephemeris[i].max() - indep_linear_ephemeris[i].min()
                )

                for j in range(i + 1, NPL):
                    # delta-freq (LS) between two planets
                    df_ij = np.abs(indep_freqs[i] - indep_freqs[j])

                    if df_ij < df_min:
                        close = True

                if close:
                    omc_freqs.append(indep_freqs[i])
                    omc_faps.append(indep_faps[i])

                else:
                    omc_freqs.append(None)
                    omc_faps.append(None)

    omc_pers = []

    for n in range(NPL):
        print("\nPLANET", n)

        # roughly model OMC based on single frequency sinusoid (if found)
        if omc_freqs[n] is not None:
            print("  periodic signal found at P =", int(1 / omc_freqs[n]), "d")

            omc_pers.append(1 / omc_freqs[n])

        else:
            print("  no sigificant periodic component found")
            omc_pers.append(
                2 * (indep_linear_ephemeris[n].max() - indep_linear_ephemeris[n].min())
            )

    # ## Determine best OMC model

    print("...running model selection routine")

    regular_transit_times = []
    full_regular_transit_times = []

    outlier_prob = []
    outlier_class = []

    for n, p in enumerate(planets):
        print(f"\nPLANET {n}")

        # grab data
        xtime = indep_linear_ephemeris[n]
        yomc = indep_transit_times[n] - indep_linear_ephemeris[n]

        # flag outliers
        if len(yomc) > 16:
            ymed = boxcar_smooth(median_filter(yomc, size=5, mode="mirror"), winsize=5)
        else:
            ymed = np.median(yomc)

        if len(yomc) > 4:
            out = np.abs(yomc - ymed) / mad_std(yomc - ymed) > 5.0
        else:
            out = np.zeros(len(yomc), dtype="bool")

        # estimate uncertainty
        yerr_expected = predict_tc_error(
            np.sqrt(p.depth), p.impact, p.duration, lcit, np.median(lc.error)
        )
        yerr_measured = indep_uncertainty[n]

        yerr = np.sqrt(yerr_expected**2 + yerr_measured**2)
        ymax = np.sqrt(mad_std(yomc) ** 2 - np.median(yerr) ** 2)

        # compare various models
        aiclist = []
        biclist = []

        if np.sum(~out) >= 16:
            min_polyorder = -1
            max_polyorder = 3
        elif np.sum(~out) >= 8:
            min_polyorder = -1
            max_polyorder = 2
        elif np.sum(~out) >= 4:
            min_polyorder = 0
            max_polyorder = 2
        else:
            min_polyorder = 1
            max_polyorder = 1

        # don't use a GP on very noisy data
        if np.median(yerr) >= 0.5 * mad_std(yomc):
            min_polyorder = np.max([0, min_polyorder])

        for polyorder in range(min_polyorder, max_polyorder + 1):
            if polyorder == -1:
                omc_model = omc.matern32_model(
                    xtime[~out], yomc[~out], yerr[~out], ymax, xtime
                )
            elif polyorder == 0:
                omc_model = omc.sin_model(
                    xtime[~out], yomc[~out], yerr[~out], omc_pers[n], xtime
                )
            elif polyorder >= 1:
                omc_model = omc.poly_model(
                    xtime[~out], yomc[~out], yerr[~out], polyorder, xtime
                )
            else:
                raise ValueError("polyorder must be >= -1")

            with omc_model:
                omc_map = omc_model.test_point
                omc_map = pmx.optimize(start=omc_map, progress=VERBOSE)
                omc_trace = pmx.sample(
                    tune=8000,
                    draws=2000,
                    chains=2,
                    target_accept=0.95,
                    start=omc_map,
                    progressbar=VERBOSE,
                )

            omc_trend = np.nanmedian(omc_trace["pred"], 0)
            residuals = yomc - omc_trend

            fig, ax = plot_omc(
                [xtime],
                [xtime + yomc],
                [xtime + omc_trend],
                out=[out],
                colors=[f"C{n}"],
            )
            if IPLOT:
                plt.show()
            else:
                plt.close()

            # calculate AIC & BIC
            npts = len(yomc)

            if polyorder == -1:
                k = np.nanmedian(omc_trace["dof"])
                print(f"keff : {k:.1f}")
            elif polyorder == 0:
                k = 3
            else:
                k = polyorder + 1

            chisq = np.sum((yomc[~out] - omc_trend[~out]) ** 2 / yerr[~out] ** 2)
            lnlike = -chisq

            aic = 2 * k - 2 * lnlike
            bic = k * np.log(npts) - 2 * lnlike

            aiclist.append(aic)
            biclist.append(bic)

            print("AIC:", np.round(aic, 1))
            print("BIC:", np.round(bic, 1))

        # choose the best model and recompute
        polyorder_aic = np.arange(min_polyorder, max_polyorder + 1)[np.argmin(aiclist)]
        polyorder_bic = np.arange(min_polyorder, max_polyorder + 1)[np.argmin(biclist)]
        polyorder = np.max([polyorder_aic, polyorder_bic])

        xt_predict = full_indep_linear_ephemeris[n]

        if polyorder == -1:
            omc_model = omc.matern32_model(
                xtime[~out], yomc[~out], yerr[~out], ymax, xt_predict
            )
        elif polyorder == 0:
            omc_model = omc.sin_model(
                xtime[~out], yomc[~out], yerr[~out], omc_pers[n], xt_predict
            )
        elif polyorder >= 1:
            omc_model = omc.poly_model(
                xtime[~out], yomc[~out], yerr[~out], polyorder, xt_predict
            )

        with omc_model:
            omc_map = omc_model.test_point
            omc_map = pmx.optimize(start=omc_map, progress=VERBOSE)
            omc_trace = pmx.sample(
                tune=8000,
                draws=2000,
                chains=2,
                target_accept=0.95,
                start=omc_map,
                progressbar=VERBOSE,
            )

        # flag outliers with K-means clustering
        omc_trend = np.nanmedian(omc_trace["pred"], 0)
        residuals = yomc - omc_trend[np.isin(xt_predict, xtime)]
        mix_model = omc.mix_model(residuals)

        with mix_model:
            mix_trace = pmx.sample(
                tune=8000, draws=2000, chains=1, target_accept=0.95, progressbar=VERBOSE
            )

        loc = np.nanmedian(mix_trace["mu"], axis=0)
        scales = np.nanmedian(1 / np.sqrt(mix_trace["tau"]), axis=0)

        fg_prob, bad = omc.flag_outliers(residuals, loc, scales)

        while np.sum(bad) / len(bad) > 0.3:
            thresh = np.max(fg_prob[bad])
            bad = fg_prob < thresh

        num_bad = np.sum(bad)
        perc_bad = 100 * num_bad / len(bad)

        err_ = np.median(yerr_measured[~bad]) * 24 * 60
        rms_ = mad_std(residuals[~bad]) * 24 * 60

        print(
            f"{num_bad:.0f} outliers found out of {len(bad):.0f} transit times ({perc_bad:.1f}%)"
        )
        print(f"measured error: {err_:.1f} min")
        print(f"residual RMS: {rms_:.1f} min")

        # save the final results
        full_omc_trend = np.nanmedian(omc_trace["pred"], 0)

        full_regular_transit_times.append(
            full_indep_linear_ephemeris[n] + full_omc_trend
        )
        regular_transit_times.append(full_regular_transit_times[n][transit_inds[n]])

        outlier_prob.append(1 - fg_prob)
        outlier_class.append(bad)

        # output a figure
        fig = plot_omc_model_selection(
            xtime,
            yomc,
            fg_prob,
            bad,
            full_indep_linear_ephemeris[n],
            full_omc_trend,
            err_,
            rms_,
            TARGET,
        )

        plt.savefig(
            os.path.join(FIGURE_DIR, TARGET + f"_ttvs_quick_{n:02d}.png"),
            bbox_inches="tight",
        )

        if IPLOT:
            plt.show()
        else:
            plt.close()

    # ## Estimate TTV scatter w/ uncertainty buffer

    ttv_scatter = np.zeros(NPL)
    ttv_buffer = np.zeros(NPL)

    for n in range(NPL):
        # estimate TTV scatter
        ttv_scatter[n] = mad_std(indep_transit_times[n] - regular_transit_times[n])

        # based on scatter in independent times, set threshold so not even one outlier is expected
        N = len(transit_inds[n])
        eta = np.max([3.0, stats.norm.interval((N - 1) / N)[1]])

        ttv_buffer[n] = eta * ttv_scatter[n] + lcit

    # ## Update TTVs

    for n, p in enumerate(planets):
        # update transit time info in Planet objects
        epoch, period = poly.polyfit(p.index, full_regular_transit_times[n], 1)

        p.epoch = np.copy(epoch)
        p.period = np.copy(period)
        p.tts = np.copy(full_regular_transit_times[n])

    print("")
    print("cumulative runtime = ", int(timer() - global_start_time), "s")
    print("")

    # # ###############################
    # # ----- FLAG OUTLIERS & RESET -----
    # # ###############################

    # ## Flag outliers based on transit model

    print("\nFlagging outliers based on transit model...\n")

    res = [None] * len(quarters)
    bad = [None] * len(quarters)

    for j, q in enumerate(quarters):
        print(f"QUARTER {q}")

        if (all_dtype[q] != "long") and (all_dtype[q] != "short"):
            pass

        else:
            if all_dtype[q] == "long":
                use = lc.quarter == q
                t_ = lc.time[use]
                f_ = lc.flux[use]
            elif all_dtype[q] == "short":
                use = sc.quarter == q
                t_ = sc.time[use]
                f_ = sc.flux[use]
            else:
                raise ValueError("cadence data type must be 'short' or 'long'")

            # grab transit times for each planet
            wp = []
            tts = []
            inds = []

            for n in range(NPL):
                rtt = regular_transit_times[n]
                use = (rtt > t_.min()) * (rtt < t_.max())

                if np.sum(use) > 0:
                    wp.append(n)
                    tts.append(rtt[use])
                    inds.append(transit_inds[n][use] - transit_inds[n][use][0])

            if len(tts) == 0:
                all_dtype[q] = all_dtype[q] + "_no_transits"
            else:
                # set up model
                starrystar = exo.LimbDarkLightCurve([U1, U2])

                orbit = exo.orbits.TTVOrbit(
                    transit_times=tts,
                    transit_inds=inds,
                    period=list(periods[wp]),
                    b=impacts[wp],
                    ror=rors[wp],
                    duration=durs[wp],
                )

                # calculate light curves
                light_curves = starrystar.get_light_curve(
                    orbit=orbit,
                    r=rors[wp],
                    t=t_,
                    oversample=oversample[q],
                    texp=texp[q],
                )

                model_flux = 1.0 + pm.math.sum(light_curves, axis=-1).eval()

                # flag outliers
                N = len(f_)
                eta = np.max([3.0, stats.norm.interval((N - 1) / N)[1]])

                res[j] = f_ - model_flux
                bad[j] = np.abs(res[j] - np.median(res[j])) / (mad_std(res[j])) > eta

                loop = 0
                count = np.sum(bad[j])
                while loop < 5:
                    bad[j] = (
                        np.abs(res[j] - np.median(res[j][~bad[j]]))
                        / (mad_std(res[j][~bad[j]]))
                        > eta
                    )

                    if np.sum(bad[j]) == count:
                        loop = 5
                    else:
                        loop += 1

            if res[j] is not None:
                print(" outliers rejected:", np.sum(bad[j]))

                plt.figure(figsize=(20, 3))
                plt.plot(t_, res[j], "k.")
                plt.plot(t_[bad[j]], res[j][bad[j]], "x", c="r", ms=20)

                if IPLOT:
                    plt.show()
                else:
                    plt.close()

    good_cadno_lc = []
    good_cadno_sc = []

    for j, q in enumerate(quarters):
        if all_dtype[q] == "long":
            use = lc.quarter == q
            good_cadno_lc.append(lc.cadno[use][~bad[j]])

        if all_dtype[q] == "short":
            use = sc.quarter == q
            good_cadno_sc.append(sc.cadno[use][~bad[j]])

    if len(good_cadno_lc) > 0:
        good_cadno_lc = np.hstack(good_cadno_lc)

    if len(good_cadno_sc) > 0:
        good_cadno_sc = np.hstack(good_cadno_sc)

    # ## Reset to raw input data

    print("\nResetting data...\n")

    if MISSION == "Kepler":
        # reset LONG CADENCE data
        lc_data = []
        for i, lcrd in enumerate(lc_raw_data):
            lc_data.append(io.LightKurve_to_LiteCurve(lcrd))

        # make sure there is at least one transit in the long cadence data
        # this shouldn't be an issue for real KOIs, but can happen for simulated data
        if np.sum(np.array(all_dtype) == "long") == 0:
            lc_data = []

        # reset SHORT CADENCE data
        sc_data = []
        for i, scrd in enumerate(sc_raw_data):
            sc_data.append(io.LightKurve_to_LiteCurve(scrd))

        # make sure there is at least one transit in the short cadence data
        # this shouldn't be an issue for real KOIs, but can happen for simulated data
        if np.sum(np.array(all_dtype) == "short") == 0:
            sc_data = []

    elif MISSION == "Kepler-Validation":
        sc_data = deepcopy(sc_raw_sim_data)
        lc_data = deepcopy(lc_raw_sim_data)

    # ## Remove flagged cadences

    for i, scd in enumerate(sc_data):
        mask = np.isin(scd.cadno, good_cadno_sc)

        if np.sum(mask) / len(mask) > 0.1:
            sc_data[i] = scd.remove_flagged_cadences(mask)

    for i, lcd in enumerate(lc_data):
        mask = np.isin(lcd.cadno, good_cadno_lc)

        if np.sum(mask) / len(mask) > 0.1:
            lc_data[i] = lcd.remove_flagged_cadences(mask)

    # # #########################
    # # ----- 2ND DETRENDING -----
    # # #########################

    print("\nDetrending lightcurves (2st pass)...\n")

    # ## Detrend the lightcurves

    # make transit masks
    for i, lcd in enumerate(lc_data):
        mask = np.zeros((NPL, len(lcd.time)), dtype="bool")

        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                lcd.time, p.tts, np.max([1 / 24, 1.5 * p.duration])
            )

        lcd.mask = mask.sum(axis=0) > 0

    for i, scd in enumerate(sc_data):
        mask = np.zeros((NPL, len(scd.time)), dtype="bool")

        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                scd.time, p.tts, np.max([1 / 24, 1.5 * p.duration])
            )

        scd.mask = mask.sum(axis=0) > 0

    # remove bad cadences
    for i, lcd in enumerate(lc_data):
        lcd = detrend.remove_bad_cadences(
            lcd, planets, rel_masksize=1.5, min_masksize=1.0 / 24
        )

    for i, scd in enumerate(sc_data):
        scd = detrend.remove_bad_cadences(
            scd, planets, rel_masksize=1.5, min_masksize=1.0 / 24
        )

    # track stellar and instrumental oscillations
    oscillation_period_by_quarter = np.ones(18) * np.nan

    for i, lcd in enumerate(lc_data):
        min_period = np.max([5 * np.max(DURS), 13 * lcit])
        oscillation_period_by_quarter[
            lcd.quarter[0]
        ] = detrend.estimate_oscillation_period(lcd, min_period=min_period)

    for i, scd in enumerate(sc_data):
        min_period = np.max([5 * np.max(DURS), 91 * scit])
        oscillation_period_by_quarter[
            scd.quarter[0]
        ] = detrend.estimate_oscillation_period(scd, min_period=min_period)

    oscillation_period_by_season = np.zeros((4, 2))

    for i in range(4):
        oscillation_period_by_season[i, 0] = np.nanmedian(
            oscillation_period_by_quarter[i::4]
        )
        oscillation_period_by_season[i, 1] = mad_std(
            oscillation_period_by_quarter[i::4], ignore_nan=True
        )

    # detrend long cadence data
    break_tolerance = np.max([int(DURS.min() / lcit * 5 / 2), 13])
    min_per = np.max([5 * np.max(DURS), 13 * lcit])

    for i, lcd in enumerate(lc_data):
        print(f"QUARTER {lcd.quarter[0]}")

        nom_per = oscillation_period_by_season[lcd.quarter[0] % 4][0]

        lcd = detrend.flatten_litecurve(
            lcd, break_tolerance, min_per, nominal_period=nom_per, verbose=VERBOSE
        )

    if len(lc_data) > 0:
        lc = detrend.stitch(lc_data)
    else:
        lc = None

    if lc is not None:
        fig = lc.plot()

        figpath = os.path.join(FIGURE_DIR, TARGET + f"_lightcurve_detrended_lc.png")
        plt.savefig(figpath, bbox_inches="tight")
        if IPLOT:
            plt.show()
        else:
            plt.close()

    # detrend short cadence data
    break_tolerance = np.max([int(DURS.min() / scit * 5 / 2), 91])
    min_per = np.max([5 * np.max(DURS), 91 * scit])

    for i, scd in enumerate(sc_data):
        print(f"QUARTER {scd.quarter[0]}")

        nom_per = oscillation_period_by_season[scd.quarter[0] % 4][0]

        scd = detrend.flatten_litecurve(
            scd, break_tolerance, min_per, nominal_period=nom_per, verbose=VERBOSE
        )

    if len(sc_data) > 0:
        sc = detrend.stitch(sc_data)
    else:
        sc = None

    if sc is not None:
        fig = sc.plot()

        figpath = os.path.join(FIGURE_DIR, TARGET + f"_lightcurve_detrended_sc.png")
        plt.savefig(figpath, bbox_inches="tight")
        if IPLOT:
            plt.show()
        else:
            plt.close()

    # # ##########################
    # # ----- FINAL DIAGNOSTICS -----
    # # ##########################

    print("\nPerforming final diagnostic checks...\n")

    # ## Make individual mask for where each planet transits

    if sc is not None:
        mask = np.zeros((NPL, len(sc.time)), dtype="bool")
        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                sc.time, p.tts, np.max([3 / 24, 1.5 * p.duration])
            )

        sc.mask = mask.sum(axis=0) > 0

    if lc is not None:
        mask = np.zeros((NPL, len(lc.time)), dtype="bool")
        for n, p in enumerate(planets):
            mask[n] = make_transit_mask(
                lc.time, p.tts, np.max([3 / 24, 1.5 * p.duration])
            )

        lc.mask = mask.sum(axis=0) > 0

    # ## Flag high quality transits (quality = 1)

    # number of photometric points expected in transit
    count_expect_lc = np.zeros(NPL, dtype="int")
    count_expect_sc = np.zeros(NPL, dtype="int")

    for n, p in enumerate(planets):
        count_expect_lc[n] = np.max([1, int(np.floor(p.duration / lcit))])
        count_expect_sc[n] = np.max([1, int(np.floor(p.duration / scit))])
        quality = np.zeros(len(p.tts), dtype="bool")

        for i, t0 in enumerate(p.tts):
            if sc is not None:
                in_sc = np.abs(sc.time - t0) / p.duration < 0.5
                near_sc = np.abs(sc.time - t0) / p.duration < 1.5

                qual_in = np.sum(in_sc) > 0.5 * count_expect_sc[n]
                qual_near = np.sum(near_sc) > 1.5 * count_expect_sc[n]

                quality[i] += qual_in * qual_near

            if lc is not None:
                in_lc = np.abs(lc.time - t0) / p.duration < 0.5
                near_lc = np.abs(lc.time - t0) / p.duration < 1.5

                qual_in = np.sum(in_lc) > 0.5 * count_expect_lc[n]
                qual_near = np.sum(near_lc) > 1.5 * count_expect_lc[n]

                quality[i] += qual_in * qual_near

        p.quality = np.copy(quality)

        if np.sum(p.quality) < 0.5 * len(p.quality):
            raise ValueError(
                f"Over 50% of transits for Planet {n} have been flagged as low quality"
            )

    # ## Flag overlapping transits

    dur_max = np.max(DURS)
    overlap = [None] * NPL

    for i in range(NPL):
        overlap[i] = np.zeros(len(planets[i].tts), dtype="bool")

        for j in range(NPL):
            if i != j:
                for ttj in planets[j].tts:
                    overlap[i] += np.abs(planets[i].tts - ttj) / dur_max < 1.5

    # ## Plot folded transit

    for n, p in enumerate(planets):
        tts = p.tts[p.quality * ~overlap[n]]

        fig = plot_folded_transit(lc, sc, tts, p.depth, p.duration, TARGET, n)
        plt.savefig(
            os.path.join(FIGURE_DIR, TARGET + f"_folded_transit_{n:02d}.png"),
            bbox_inches="tight",
        )
        if IPLOT:
            plt.show()
        else:
            plt.close()

    # # ###################
    # # ----- SAVE & EXIT -----
    # # ###################

    # ## Save nominal transit parameters

    print("\nSaving nominal transit parameters...\n")

    transit_parameters = io.transit_parameters_to_dataframe(
        KOI_ID, KIC_ID, planets, [U1, U2]
    )
    transit_parameters.to_csv(
        os.path.join(RESULTS_DIR, f"{TARGET}_transit_parameters.csv")
    )

    # ## Save transit times

    print("\nSaving transit times...\n")

    for n, p in enumerate(planets):
        keep = np.isin(regular_transit_times[n], p.tts[p.quality])

        data_out = np.vstack(
            [
                transit_inds[n][keep],
                indep_transit_times[n][keep],
                regular_transit_times[n][keep],
                outlier_prob[n][keep],
                outlier_class[n][keep],
            ]
        ).swapaxes(0, 1)

        fname_out = os.path.join(RESULTS_DIR, f"{TARGET}_{n:02d}_quick.ttvs")
        np.savetxt(
            fname_out,
            data_out,
            fmt=("%1d", "%.8f", "%.8f", "%.8f", "%1d"),
            delimiter="\t",
        )

    # ## Save detrended lightcurves

    print("\nSaving detrended lightcurves...\n")

    if lc is not None:
        filename = os.path.join(RESULTS_DIR, f"{TARGET}_lc_detrended.fits")
        lc.to_fits(TARGET, filename, cadence="LONG")
    else:
        print("  No long cadence data")

    if sc is not None:
        filename = os.path.join(RESULTS_DIR, f"{TARGET}_sc_detrended.fits")
        sc.to_fits(TARGET, filename, cadence="SHORT")
    else:
        print("  No short cadence data")

    # ## Save stellar oscillation periods

    print("\nSaving stellar oscillation periods...\n")

    filename = os.path.join(RESULTS_DIR, f"{TARGET}_stellar_oscillations.txt")
    np.savetxt(
        filename,
        np.stack([np.arange(18), oscillation_period_by_quarter]).T,
        fmt=["%d", "%.9e"],
    )

    # ## Exit program

    print("")
    print("+" * shutil.get_terminal_size().columns)
    print(
        f"Automated lightcurve detrending complete {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}"
    )
    print(f"Total runtime = {(timer()-global_start_time)/60:.1f} min")
    print("+" * shutil.get_terminal_size().columns)


if __name__ == "__main__":
    main()



__all__ = ['Ephemeris']

from copy import deepcopy
import numpy as np
from scipy.interpolate import CubicSpline
import warnings


class Ephemeris:
    """Ephemeris
    """
    def __init__(self, 
                 period=None,
                 epoch=None,
                 index=None, 
                 ttime=None, 
                 error=None,
                 quality=None,
                 t_min=None,
                 t_max=None
                ):
        
        # check inputs
        init_linear_ephem = (
            (period is not None) & 
            (epoch is not None) &
            (t_min is not None) &
            (t_max is not None)
        )

        init_ttv_ephem = (
            (index is not None) &
            (ttime is not None)
        )

        if init_linear_ephem + init_ttv_ephem != 1:
            raise ValueError("must supply exactly one of (ttime, index) or (period, epoch)")
        
        # Case 1 : linear ephemeris
        if init_linear_ephem:
            self.period = period
            self.epoch = epoch
            self.t_min = t_min
            self.t_max = t_max

            self._adjust_epoch(self.t_min)

            self.ttime = np.arange(self.epoch, t_max, self.period)
            self.index = np.array(np.round((self.ttime - self.epoch) / self.period), dtype=int)
            self.error = None
            self.quality = None

        # Case 2 : TTV ephemeris
        elif init_ttv_ephem:
            self.index = index
            self.ttime = ttime
            self.error = error
            self.quality = quality

            self.period, self.epoch = self.fit_linear_ephemeris()

            if t_min is not None:
                self.t_min = t_min
            else:
                self.t_min = self.ttime.min() - self.period / 2
            if t_max is not None:
                self.t_max = t_max
            else:
                self.t_max = self.ttime.max() + self.period / 2

        # clip vectors to range (t_min, t_max)
        use = (self.ttime >= self.t_min) & (self.ttime <= self.t_max)
        for k in self.__dict__.keys():
            if type(self.__dict__[k]) is np.ndarray:
                self.__setattr__(k, self.__dict__[k][use])

        # put epoch in range (t_min, t_min + period)
        self._adjust_epoch(self.t_min)

        # set static period, epoch, and ephemeris
        self._set_static_references()


    def _adjust_epoch(self, t_min):
        """
        Put epoch in range (t_min, t_min + period)

        Args:
          t_min (float) : minimum time

        Returns:
          Ephemeris : self
        """
        if self.epoch < t_min:
            adj = 1 + (t_min - self.epoch) // self.period
            self.epoch += adj * self.period
        if self.epoch > (t_min + self.period):
            adj = (self.epoch - t_min) // self.period
            self.epoch -= adj * self.period
    
    
    def _set_static_references(self):
        assert not hasattr(self, '_static_period'), "attribute '_static_period' already exists"
        self._static_period = np.copy(self.period)

        assert not hasattr(self, '_static_epoch'), "attribute '_static_epoch' already exists"
        self._static_epoch = np.copy(self.epoch)


    def fit_linear_ephemeris(self):
        """
        Fit a linear ephmeris using unweighted least squares
        """
        A = np.ones((len(self.index), 2))
        A[:, 0] = self.index

        return np.linalg.lstsq(A, self.ttime, rcond=None)[0]
    

    def eval_linear_ephemeris(self, index=None):
        """
        Calculate linear ephemeris from period and epoch

        Returns:
          ndarray : transit times according to linear ephemeris
        """
        if index is None:
            index = self.index

        return self.epoch + self.period*index
    

    def update_period_and_epoch(self):
        """
        Recompute linear ephemeris to ensure consistency

        Returns:
          Ephemeris : self
        """
        self.period, self.epoch = self.fit_linear_ephemeris()

        return self
        

    def interpolate(self, method, full=False):
        """
        Interpolate poor quality transit times and optionally interpolate missing transit times

        Args:
          method (str) : 'linear' or 'spline'
          full (bool) : True to interpolate missing transit times (default=False)

        Returns:
          Ephemeris : self
        """        
        if self.quality is not None:
            q = self.quality
        else:
            q = np.ones(len(self.ttime), dtype=bool)

        transit_exists = deepcopy(self.index)
        index_full = np.arange(self.index.min(), self.index.max()+1, dtype=int)

        if method == 'linear':
            ttime_full = self.epoch + self.period*index_full
        
        elif method == 'spline':
            spline = CubicSpline(self.index[q],
                                self.ttime[q], 
                                extrapolate=True,
                                bc_type='natural'
                                )

            ttime_full = spline(index_full)

        ttime_full[transit_exists[q]] = self.ttime[q]

        error_full = np.zeros(len(ttime_full))*np.nan
        if self.error is not None:
            error_full[transit_exists[q]] = self.error[q]

        quality_full = np.ones(len(ttime_full), dtype=bool)
        if self.quality is not None:
            quality_full[transit_exists] = self.quality
        
        if full:
            keep = index_full
        else:
            keep = transit_exists
        
        self.index = index_full[keep]
        self.ttime = ttime_full[keep]
        self.error = error_full[keep]
        self.quality = quality_full[keep]

        self = self.update_period_and_epoch()

        return self
    
    
    def update_from_omc(self, omc):
        assert np.all(np.isclose(self._static_period, omc._static_period, rtol=1e-12)), "static periods do not match"
        assert np.all(np.isclose(self._static_epoch, omc._static_epoch, rtol=1e-12)), "static epochs do not match"

        if len(self.ttime) != len(omc.xtime):
            warnings.warn(f"updated ephemeris has {len(omc.xtime)} transit times (was {len(self.ttimes)})")
        
        self.index = omc.index.copy()
        self.ttime = self._static_epoch + self._static_period*self.index + omc.ymod
        self.error = omc.yerr.copy()
        self.quality = omc.quality.copy()

        self.period, self.epoch = self.fit_linear_ephemeris()

        return self


import os
import sys

import numpy as np
from pathlib import Path
import warnings
from alderaan.ephemeris import Ephemeris
from alderaan.planet import Planet
from alderaan.utils.io import parse_koi_catalog

warnings.simplefilter('always', UserWarning)


base_path = Path(__file__).resolve().parents[1]
catalog_csv = os.path.join(base_path, 'alderaan/examples/catalogs/kepler_dr25_gaia_dr2_crossmatch.csv')

koi_id = 'K00137'   # Kepler-18
catalog = parse_koi_catalog(catalog_csv, koi_id)

assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

t_min = 0.
t_max = 1400.

NPL = int(catalog.npl[0])
planets = [None]*NPL

for n in range(NPL):
    planets[n] = Planet(catalog, koi_id, n)

for n, p in enumerate(planets):
    ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
    p = p.update_ephemeris(ephemeris)

print("\npassing")